from datetime import date

from sqlalchemy import Boolean, Column, Date, Integer, String

from mavedb.db.base import Base


class License(Base):
    __tablename__ = "licenses"

    id = Column(Integer, primary_key=True)
    long_name = Column(String, nullable=False, unique=True)
    short_name = Column(String, nullable=False, unique=True)
    text = Column(String, nullable=False, unique=False)
    link = Column(String, nullable=True, unique=False)
    version = Column(String, nullable=True, unique=False)
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)
    active = Column(Boolean, nullable=False)



import datetime
from urllib.parse import urlparse

import idutils

from mavedb.lib.validation.constants.publication import valid_dbnames
from mavedb.lib.validation.exceptions import ValidationError


def validate_db_name(db_name: str):
    if db_name.strip() == "" or not db_name:
        raise ValidationError("db_name should not be empty")
    if db_name not in valid_dbnames:
        raise ValidationError(
            f"The `db_name` key within the identifier attribute of the external "
            f"publication identifier should take one of the following values: "
            f"{valid_dbnames}."
        )


def identifier_valid_for(identifier: str) -> dict[str, bool]:
    """
    Returns a list of db_names for which the provided identifier is
    a valid identifier.

    Parameters
    __________
    identifier : str
        The identifier to check

    Returns
    _______
    dict[str, bool]
        The db_names for which this identifier is a valid identifier
    """
    return {
        "PubMed": validate_pubmed(identifier),
        "bioRxiv": validate_biorxiv(identifier),
        "medRxiv": validate_medrxiv(identifier),
        "Crossref": idutils.is_doi(identifier) is not None,
    }


def infer_identifier_from_url(identifier: str) -> str:
    """
    Infers an identifier from a potential URL based on the database we believe the URL
    to be from.

    Parameters
    __________
    identifier : str
        The identifier / URL to parse

    Returns
    _______
    str
        The parsed identifier from the url or the original identifier.
    """
    url = urlparse(identifier)
    if url.netloc:
        # http://www.dx.doi.org/{DOI}
        if "dx.doi.org" in url.netloc:
            identifier = url.path.strip("/")

        # https://www.biorxiv.org/content/10.1101/2024.04.26.591310, # https://www.medrxiv.org/content/10.1101/2024.04.26.59131023
        elif "biorxiv.org" in url.netloc or "medrxiv.org" in url.netloc:
            identifier = url.path.strip("/").split("/")[-1]

        # https://pubmed.ncbi.nlm.nih.gov/24567513/, http://www.ncbi.nlm.nih.gov/pubmed/432
        elif "ncbi.nlm.nih.gov" in url.netloc:
            identifier = url.path.strip("/").split("/")[-1]

        # The url does not come from an accepted database.
        else:
            return identifier

    return identifier


def validate_publication(identifier: str) -> None:
    """
    Validates that a passed identifier is one we accept. Currently allowed
    publication identifiers:

    - PMIDs: Pubmed IDs are an 8 character integer (no leading zeros)
    - bioRxiv IDs: biorXiv IDs are either 6 digit identifiers if the
                   preprint article is from before 2019.12.11 (123456)
                   or the article publication date with a 6 digit
                   suffix if published after that date (2019.12.12.123456).
    - medRxiv IDs: medRxiv IDs are either 8 digit identifiers if the
                   preprint article is from before 2019.12.11 (12345678)
                   or the article publication date with a 8 digit
                   suffix if published after that date (2019.12.12.12345678).
    - valid DOIs: We use a RegEx provided by IDUtils to check this constraint.

    Note that preprint identifiers may have leading zeros, while PMIDs may not.

    Parameters
    __________
    identifier : str
        The identifier to be validated.

    Returns
    _______
    None
        If the identifier is in one of the formats we accept.
        NOTE: This does not imply that the identifier exists
        in the Crossref, PubMed, bioRxiv, or medRxiv databases.

    Raises
    ______
    ValidationError
        If the identifier is not an accepted publication identifier.
    """
    if not (
        validate_pubmed(identifier)
        or validate_biorxiv(identifier)
        or validate_medrxiv(identifier)
        or idutils.is_doi(identifier)
    ):
        raise ValidationError(f"'{identifier}' is not a valid DOI or a valid PubMed, bioRxiv, or medRxiv identifier.")


def validate_pubmed(identifier: str) -> bool:
    """
    Validates whether the provided identifier is a valid PubMed identifier.

    Parameters
    __________
    identifier : str
        The identifier to validate

    Returns
    _______
    bool
        Whether the identifier is valid.

    """
    return idutils.is_pmid(identifier) is not None


def validate_biorxiv(identifier: str) -> bool:
    """
    Validates whether the provided identifier is a valid biorxiv identifier.

    Parameters
    __________
    identifier : str
        The identifier to validate

    Returns
    _______
    bool
        Whether the identifier is valid.

    Raises
    ______
    ValueError (via _validate_new_preprint_format)
        If the provided date_parts cannot be coerced to a `date` object.
    """
    if "." in identifier:
        return _validate_new_preprint_format(identifier, 6)

    if identifier.isdigit():
        return _validate_old_preprint_format(identifier, 6)

    return False


def validate_medrxiv(identifier: str) -> bool:
    """
    Validates whether the provided identifier is a valid medrxiv identifier.

    Parameters
    __________
    identifier : str
        The identifier to validate.

    Returns
    _______
    bool
        Whether the identifier is valid.

    Raises
    ______
    ValueError (via _validate_new_preprint_format)
        If the provided date_parts cannot be coerced to a `date` object.
    """
    if "." in identifier:
        return _validate_new_preprint_format(identifier, 8)

    if identifier.isdigit():
        return _validate_old_preprint_format(identifier, 8)

    return False


def _validate_new_preprint_format(identifier: str, suffix_len: int) -> bool:
    """
    Validates whether the provided identifier is a valid pre-print identifier
    in the new format.

    Parameters
    __________
    identifier : str
        The identifier to validate.

    suffix_len : int
        The desired suffix length of the identifier.

    Returns
    _______
    bool
        Whether the identifier is valid.

    Raises
    ______
    ValueError (via _date_is_after_preprint_changeover)
        If the provided date_parts cannot be coerced to a `date` object.
    """
    preprint_id = identifier.split(".")
    preprint_suffix = preprint_id[-1]
    preprint_date = tuple(preprint_id[:-1])

    if not preprint_suffix.isdigit():
        return False

    if len(preprint_suffix) == suffix_len:
        return len(preprint_date) == 3 and _date_is_after_preprint_changeover(preprint_date)

    return False


def _validate_old_preprint_format(identifier: str, suffix_len: int) -> bool:
    """
    Validates that a passed identifier is an old pre-print (bioRxiv or medRxiv)
    identifier.

    Parameters
    __________
    identifier : str
        The identifier to validate.

    suffix_len : int
        The desired suffix length of the identifier.

    Returns
    _______
    bool
        Whether the identifier is valid.
    """
    return len(identifier) == suffix_len and identifier.isdigit()


def _date_is_after_preprint_changeover(date_parts: tuple[str, str, str]) -> bool:
    """
    Checks if a passed date with date parts comprised ot a three-tuple occurred
    after pre-print servers began using a new identifier format.

    Parameters
    __________
    date_parts : tuple[str, str, str]
        A tuple with three string entries (year, month, day) to check against
        the changeover date.

    Returns
    _______
    bool
        Whether the date occured after the pre-print server identifier changeover.

    Raises
    ______
    ValueError (via datetime.datetime.strptime)
        If the passed date_parts cannot be coerced to a `date` object.
    """
    date_cutoff = datetime.datetime.strptime("20191211", "%Y%m%d").date()
    date_part = datetime.datetime.strptime("".join(date_parts), "%Y%m%d").date()

    if date_part > date_cutoff:
        return True
    else:
        return False



from enum import Enum


class LogType(str, Enum):
    api_request = "api_request"
    worker_job = "worker_job"


class Source(str, Enum):
    docs = "docs"
    other = "other"
    web = "web"
    worker = "worker"



import logging
from typing import Optional

from sqlalchemy import func, or_
from sqlalchemy.orm import Session

from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.models.contributor import Contributor
from mavedb.models.controlled_keyword import ControlledKeyword
from mavedb.models.experiment import Experiment
from mavedb.models.experiment_controlled_keyword import ExperimentControlledKeywordAssociation
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.models.score_set import ScoreSet
from mavedb.models.user import User
from mavedb.view_models.search import ExperimentsSearch

logger = logging.getLogger(__name__)


def search_experiments(
    db: Session, owner_or_contributor: Optional[User], search: ExperimentsSearch
) -> list[Experiment]:
    save_to_logging_context({"experiment_search_criteria": search.dict()})

    query = db.query(Experiment)
    # .filter(ScoreSet.private.is_(False))

    if owner_or_contributor is not None:
        query = query.filter(
            or_(
                Experiment.created_by_id == owner_or_contributor.id,
                Experiment.contributors.any(Contributor.orcid_id == owner_or_contributor.username),
            )
        )

    if search.published is not None:
        if search.published:
            query = query.filter(Experiment.published_date.isnot(None))
        else:
            query = query.filter(Experiment.published_date.is_(None))

    if search.text:
        lower_search_text = search.text.lower()
        query = query.filter(
            or_(
                Experiment.urn.icontains(lower_search_text),
                Experiment.title.icontains(lower_search_text),
                Experiment.short_description.icontains(lower_search_text),
                Experiment.abstract_text.icontains(lower_search_text),
                Experiment.publication_identifiers.any(
                    func.lower(PublicationIdentifier.identifier).icontains(lower_search_text)
                ),
                Experiment.publication_identifiers.any(
                    func.lower(PublicationIdentifier.abstract).icontains(lower_search_text)
                ),
                Experiment.publication_identifiers.any(
                    func.lower(PublicationIdentifier.title).icontains(lower_search_text)
                ),
                Experiment.publication_identifiers.any(
                    func.lower(PublicationIdentifier.publication_journal).icontains(lower_search_text)
                ),
                Experiment.publication_identifiers.any(
                    func.jsonb_path_exists(
                        PublicationIdentifier.authors,
                        f"""$[*].name ? (@ like_regex "{lower_search_text}" flag "i")""",
                    )
                ),
            )
        )

    if search.publication_identifiers:
        query = query.filter(
            ScoreSet.publication_identifiers.any(PublicationIdentifier.identifier.in_(search.publication_identifiers))
        )

    if search.authors:
        query = query.filter(
            Experiment.publication_identifiers.any(
                func.jsonb_path_query_array(PublicationIdentifier.authors, "$.name").op("?|")(search.authors)
            )
        )

    if search.databases:
        query = query.filter(
            Experiment.publication_identifiers.any(PublicationIdentifier.db_name.in_(search.databases))
        )

    if search.journals:
        query = query.filter(
            Experiment.publication_identifiers.any(PublicationIdentifier.publication_journal.in_(search.journals))
        )

    if search.keywords:
        query = query.filter(
            Experiment.keyword_objs.any(
                ExperimentControlledKeywordAssociation.controlled_keyword.has(
                    ControlledKeyword.value.in_(search.keywords)
                )
            )
        )

    items: list[Experiment] = query.order_by(Experiment.urn, Experiment.title).all()
    if not items:
        items = []

    save_to_logging_context({"matching_resources": len(items)})
    logger.debug(msg="Experiment search yielded {len(items)} matching resources.", extra=logging_context())

    return items



from typing import Any, Optional

from fastapi import APIRouter, Depends
from fastapi.exceptions import HTTPException
from sqlalchemy.exc import MultipleResultsFound
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.models.mapped_variant import MappedVariant
from mavedb.models.variant import Variant
from mavedb.view_models import mapped_variant


async def fetch_mapped_variant_by_variant_urn(db, urn: str) -> Optional[MappedVariant]:
    """
    We may combine this function back to show_mapped_variant if none of any new function call it.
    Fetch one mapped variant by variant URN.

    :param db: An active database session.
    :param urn: The variant URN.
    :return: The mapped variant, or None if the URL was not found or refers to a private score set not owned by the specified
        user.
    """
    try:
        item = (
            db.query(MappedVariant)
            .filter(Variant.urn == urn)
            .filter(MappedVariant.variant_id == Variant.id)
            .filter(MappedVariant.current is True)
            .one_or_none()
        )
    except MultipleResultsFound:
        raise HTTPException(status_code=500, detail=f"Multiple variants with URN {urn} were found.")
    if not item:
        raise HTTPException(status_code=404, detail=f"Mapped variant with URN {urn} not found")
    return item


router = APIRouter(
    prefix="/api/v1/mapped-variants", tags=["mapped variants"], responses={404: {"description": "Not found"}}
)


@router.get("/{urn}", status_code=200, response_model=mapped_variant.MappedVariant, responses={404: {}, 500: {}})
async def show_mapped_variant(*, urn: str, db: Session = Depends(deps.get_db)) -> Any:
    """
    Fetch a mapped variant by URN.
    """

    return await fetch_mapped_variant_by_variant_urn(db, urn)


# for testing only
# @router.post("/map/{urn}", status_code=200, responses={404: {}, 500: {}})
# async def map_score_set(*, urn: str, worker: ArqRedis = Depends(deps.get_worker)) -> Any:
#     await worker.lpush(MAPPING_QUEUE_NAME, urn)  # type: ignore
#     await worker.enqueue_job(
#         "variant_mapper_manager",
#         None,
#         None,
#         None
#     )



from sqlalchemy import Column, ForeignKey, Integer
from sqlalchemy.orm import Mapped, backref, relationship

from mavedb.db.base import Base
from mavedb.models.ensembl_identifier import EnsemblIdentifier
from mavedb.models.target_gene import TargetGene


class EnsemblOffset(Base):
    __tablename__ = "ensembl_offsets"

    identifier_id = Column(Integer, ForeignKey("ensembl_identifiers.id"), nullable=False, primary_key=True)
    identifier: Mapped[EnsemblIdentifier] = relationship(
        "EnsemblIdentifier", backref=backref("target_gene_offsets", uselist=True)
    )
    target_gene_id = Column(Integer, ForeignKey("target_genes.id"), nullable=False, primary_key=True)
    target_gene: Mapped[TargetGene] = relationship(back_populates="ensembl_offset")

    offset = Column(Integer, nullable=False)



import logging
import click
from datetime import date
from typing import Sequence, Optional

from sqlalchemy import cast, select
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Session

from mavedb.data_providers.services import vrs_mapper
from mavedb.lib.logging.context import format_raised_exception_info_as_dict
from mavedb.models.enums.mapping_state import MappingState
from mavedb.models.score_set import ScoreSet
from mavedb.models.mapped_variant import MappedVariant
from mavedb.models.target_gene import TargetGene
from mavedb.models.variant import Variant

from mavedb.scripts.environment import script_environment, with_database_session

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def variant_from_mapping(db: Session, mapping: dict, dcd_mapping_version: str) -> MappedVariant:
    variant_urn = mapping.get("mavedb_id")
    variant = db.scalars(select(Variant).where(Variant.urn == variant_urn)).one()

    return MappedVariant(
        variant_id=variant.id,
        pre_mapped=mapping.get("pre_mapped"),
        post_mapped=mapping.get("post_mapped"),
        modification_date=date.today(),
        mapped_date=date.today(),  # since this is a one-time script, assume mapping was done today
        vrs_version=mapping.get("vrs_version"),
        mapping_api_version=dcd_mapping_version,
        error_message=mapping.get("error_message"),
        current=True,
    )


@script_environment.command()
@with_database_session
@click.argument("urns", nargs=-1)
@click.option("--all", help="Populate mapped variants for every score set in MaveDB.", is_flag=True)
def populate_mapped_variant_data(db: Session, urns: Sequence[Optional[str]], all: bool):
    score_set_ids: Sequence[Optional[int]]
    if all:
        score_set_ids = db.scalars(select(ScoreSet.id)).all()
        logger.info(
            f"Command invoked with --all. Routine will populate mapped variant data for {len(urns)} score sets."
        )
    else:
        score_set_ids = db.scalars(select(ScoreSet.id).where(ScoreSet.urn.in_(urns))).all()
        logger.info(f"Populating mapped variant data for the provided score sets ({len(urns)}).")

    vrs = vrs_mapper()

    for idx, ss_id in enumerate(score_set_ids):
        if not ss_id:
            continue

        score_set = db.scalar(select(ScoreSet).where(ScoreSet.id == ss_id))
        if not score_set:
            logger.warning(f"Could not fetch score set with id={ss_id}.")
            continue

        try:
            existing_mapped_variants = (
                db.query(MappedVariant).join(Variant).join(ScoreSet).filter(MappedVariant.current.is_(True)).all()
            )

            for variant in existing_mapped_variants:
                variant.current = False

            assert score_set.urn
            logger.info(f"Mapping score set {score_set.urn}.")
            mapped_scoreset = vrs.map_score_set(score_set.urn)
            logger.info(f"Done mapping score set {score_set.urn}.")

            dcd_mapping_version = mapped_scoreset["dcd_mapping_version"]
            mapped_scores = mapped_scoreset.get("mapped_scores")

            if not mapped_scores:
                # if there are no mapped scores, the score set failed to map.
                score_set.mapping_state = MappingState.failed
                score_set.mapping_errors = {"error_message": mapped_scoreset.get("error_message")}
                db.commit()
                logger.info(f"No mapped variants available for {score_set.urn}.")
            else:
                computed_genomic_ref = mapped_scoreset.get("computed_genomic_reference_sequence")
                mapped_genomic_ref = mapped_scoreset.get("mapped_genomic_reference_sequence")
                computed_protein_ref = mapped_scoreset.get("computed_protein_reference_sequence")
                mapped_protein_ref = mapped_scoreset.get("mapped_protein_reference_sequence")

                # assumes one target gene per score set, which is currently true in mavedb as of sept. 2024.
                target_gene = db.scalars(
                    select(TargetGene)
                    .join(ScoreSet)
                    .where(
                        ScoreSet.urn == str(score_set.urn),
                    )
                ).one()

                excluded_pre_mapped_keys = {"sequence"}
                if computed_genomic_ref and mapped_genomic_ref:
                    pre_mapped_metadata = computed_genomic_ref
                    target_gene.pre_mapped_metadata = cast(
                        {
                            "genomic": {
                                k: pre_mapped_metadata[k]
                                for k in set(list(pre_mapped_metadata.keys())) - excluded_pre_mapped_keys
                            }
                        },
                        JSONB,
                    )
                    target_gene.post_mapped_metadata = cast({"genomic": mapped_genomic_ref}, JSONB)
                elif computed_protein_ref and mapped_protein_ref:
                    pre_mapped_metadata = computed_protein_ref
                    target_gene.pre_mapped_metadata = cast(
                        {
                            "protein": {
                                k: pre_mapped_metadata[k]
                                for k in set(list(pre_mapped_metadata.keys())) - excluded_pre_mapped_keys
                            }
                        },
                        JSONB,
                    )
                    target_gene.post_mapped_metadata = cast({"protein": mapped_protein_ref}, JSONB)
                else:
                    raise ValueError(f"incomplete or inconsistent metadata for score set {score_set.urn}")

                mapped_variants = [
                    variant_from_mapping(db=db, mapping=mapped_score, dcd_mapping_version=dcd_mapping_version)
                    for mapped_score in mapped_scores
                ]
                logger.debug(f"Done constructing {len(mapped_variants)} mapped variant objects.")

                num_successful_variants = len(
                    [variant for variant in mapped_variants if variant.post_mapped is not None]
                )
                logger.debug(
                    f"{num_successful_variants}/{len(mapped_variants)} variants generated a post-mapped VRS object."
                )

                if num_successful_variants == 0:
                    score_set.mapping_state = MappingState.failed
                    score_set.mapping_errors = {"error_message": "All variants failed to map"}
                elif num_successful_variants < len(mapped_variants):
                    score_set.mapping_state = MappingState.incomplete
                else:
                    score_set.mapping_state = MappingState.complete

                db.bulk_save_objects(mapped_variants)
                db.commit()
                logger.info(f"Done populating {len(mapped_variants)} mapped variants for {score_set.urn}.")

        except Exception as e:
            logging_context = {
                "mapped_score_sets": urns[:idx],
                "unmapped_score_sets": urns[idx:],
            }
            logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
            logger.error(f"Score set {score_set.urn} failed to map.", extra=logging_context)
            logger.info(f"Rolling back all changes for scoreset {score_set.urn}")
            db.rollback()

        logger.info(f"Done with score set {score_set.urn}. ({idx+1}/{len(urns)}).")

    logger.info("Done populating mapped variant data.")


if __name__ == "__main__":
    populate_mapped_variant_data()



import logging
from typing import Any

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import UserData
from mavedb.lib.authorization import RoleRequirer, require_current_user
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.permissions import Action, assert_permission
from mavedb.models.enums.user_role import UserRole
from mavedb.models.user import User
from mavedb.view_models import user

router = APIRouter(
    prefix="/api/v1", tags=["access keys"], responses={404: {"description": "Not found"}}, route_class=LoggedRoute
)

logger = logging.getLogger(__name__)


# Trailing slash is deliberate
@router.get("/users/", status_code=200, response_model=list[user.AdminUser], responses={404: {}})
async def list_users(
    *, db: Session = Depends(deps.get_db), user_data: UserData = Depends(RoleRequirer([UserRole.admin]))
) -> Any:
    """
    List users.
    """
    items = db.query(User).order_by(User.username).all()
    return items


@router.get("/users/me", status_code=200, response_model=user.CurrentUser, responses={404: {}, 500: {}})
async def show_me(*, user_data: UserData = Depends(require_current_user)) -> Any:
    """
    Return the current user.
    """
    return user_data.user


@router.get("/users/{id}", status_code=200, response_model=user.AdminUser, responses={404: {}, 500: {}})
async def show_user(
    *, id: int, user_data: UserData = Depends(RoleRequirer([UserRole.admin])), db: Session = Depends(deps.get_db)
) -> Any:
    """
    Fetch a single user by ID.
    """
    save_to_logging_context({"requested_user": id})
    item = db.query(User).filter(User.id == id).one_or_none()
    if not item:
        logger.warning(msg="Could not show user; Requested user does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"User with ID {id} not found")
    return item


@router.put("/users/me", status_code=200, response_model=user.CurrentUser, responses={404: {}, 500: {}})
async def update_me(
    *,
    user_update: user.CurrentUserUpdate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Update the current user.
    """
    current_user = user_data.user
    assert_permission(user_data, current_user, Action.UPDATE)
    current_user.email = user_update.email
    current_user.is_first_login = False
    db.add(current_user)
    db.commit()
    db.refresh(current_user)
    return current_user


@router.put("/users/me/has-logged-in", status_code=200, response_model=user.CurrentUser, responses={404: {}, 500: {}})
async def user_has_logged_in(
    *,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Update the current users log in state.
    """
    current_user = user_data.user
    assert_permission(user_data, current_user, Action.UPDATE)
    current_user.is_first_login = False
    db.add(current_user)
    db.commit()
    db.refresh(current_user)
    return current_user


# Double slash is deliberate.
@router.put("/users//{id}", status_code=200, response_model=user.AdminUser, responses={404: {}, 500: {}})
async def update_user(
    *,
    id: int,
    item_update: user.AdminUserUpdate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Update a user.
    """
    save_to_logging_context({"requested_user": id})
    item = db.query(User).filter(User.id == id).one_or_none()
    if not item:
        logger.warning(msg="Could not update user; Requested user does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"User with id {id} not found.")

    assert_permission(user_data, item, Action.UPDATE)
    assert_permission(user_data, item, Action.ADD_ROLE)

    if item_update.first_name:
        item.first_name = item_update.first_name
    if item_update.last_name:
        item.last_name = item_update.last_name
    if item_update.email:
        item.email = item_update.email
    if item_update.roles:
        await item.set_roles(db, item_update.roles)

    db.add(item)
    db.commit()
    db.refresh(item)
    return item



import asyncio
import functools
import logging
from contextlib import asynccontextmanager
from datetime import date, timedelta
from typing import Any, Optional

import pandas as pd
from arq import ArqRedis
from arq.jobs import Job, JobStatus
from cdot.hgvs.dataproviders import RESTDataProvider
from sqlalchemy import cast, delete, null, select
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Session

from mavedb.data_providers.services import vrs_mapper
from mavedb.lib.exceptions import MappingEnqueueError, NonexistentMappingReferenceError, NonexistentMappingResultsError
from mavedb.lib.logging.context import format_raised_exception_info_as_dict
from mavedb.lib.score_sets import (
    columns_for_dataset,
    create_variants,
    create_variants_data,
)
from mavedb.lib.slack import send_slack_message
from mavedb.lib.validation.dataframe import (
    validate_and_standardize_dataframe_pair,
)
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.models.enums.mapping_state import MappingState
from mavedb.models.enums.processing_state import ProcessingState
from mavedb.models.mapped_variant import MappedVariant
from mavedb.models.score_set import ScoreSet
from mavedb.models.user import User
from mavedb.models.variant import Variant

logger = logging.getLogger(__name__)

MAPPING_QUEUE_NAME = "vrs_mapping_queue"
MAPPING_CURRENT_ID_NAME = "vrs_mapping_current_job_id"
BACKOFF_LIMIT = 5
BACKOFF_IN_SECONDS = 15


@asynccontextmanager
async def mapping_in_execution(redis: ArqRedis, job_id: str):
    await redis.set(MAPPING_CURRENT_ID_NAME, job_id)
    try:
        yield
    finally:
        await redis.set(MAPPING_CURRENT_ID_NAME, "")


def setup_job_state(ctx, invoker: int, resource: Optional[str], correlation_id: str):
    ctx["state"][ctx["job_id"]] = {
        "application": "mavedb-worker",
        "user": invoker,
        "resource": resource,
        "correlation_id": correlation_id,
    }
    return ctx["state"][ctx["job_id"]]


async def enqueue_job_with_backoff(
    redis: ArqRedis, job_name: str, attempt: int, *args
) -> tuple[Optional[str], bool, Any]:
    new_job_id = None
    backoff = None
    limit_reached = attempt > BACKOFF_LIMIT
    if not limit_reached:
        limit_reached = True
        backoff = BACKOFF_IN_SECONDS * (2**attempt)
        attempt = attempt + 1

        # NOTE: for jobs supporting backoff, `attempt` should be the final argument.
        new_job = await redis.enqueue_job(
            job_name,
            *args,
            attempt,
            _defer_by=timedelta(seconds=backoff),
        )

        if new_job:
            new_job_id = new_job.job_id

    return (new_job_id, not limit_reached, backoff)


async def create_variants_for_score_set(
    ctx, correlation_id: str, score_set_id: int, updater_id: int, scores: pd.DataFrame, counts: pd.DataFrame
):
    """
    Create variants for a score set. Intended to be run within a worker.
    On any raised exception, ensure ProcessingState of score set is set to `failed` prior
    to exiting.
    """
    logging_context = {}
    try:
        db: Session = ctx["db"]
        hdp: RESTDataProvider = ctx["hdp"]
        redis: ArqRedis = ctx["redis"]
        score_set = db.scalars(select(ScoreSet).where(ScoreSet.id == score_set_id)).one()

        logging_context = setup_job_state(ctx, updater_id, score_set.urn, correlation_id)
        logger.info(msg="Began processing of score set variants.", extra=logging_context)

        updated_by = db.scalars(select(User).where(User.id == updater_id)).one()

        score_set.modified_by = updated_by
        score_set.processing_state = ProcessingState.processing
        score_set.mapping_state = MappingState.pending_variant_processing
        logging_context["processing_state"] = score_set.processing_state.name
        logging_context["mapping_state"] = score_set.mapping_state.name

        db.add(score_set)
        db.commit()
        db.refresh(score_set)

        if not score_set.target_genes:
            logger.warning(
                msg="No targets are associated with this score set; could not create variants.",
                extra=logging_context,
            )
            raise ValueError("Can't create variants when score set has no targets.")

        validated_scores, validated_counts = validate_and_standardize_dataframe_pair(
            scores, counts, score_set.target_genes, hdp
        )

        score_set.dataset_columns = {
            "score_columns": columns_for_dataset(validated_scores),
            "count_columns": columns_for_dataset(validated_counts),
        }

        # Delete variants after validation occurs so we don't overwrite them in the case of a bad update.
        if score_set.variants:
            existing_variants = db.scalars(select(Variant.id).where(Variant.score_set_id == score_set.id)).all()
            db.execute(delete(MappedVariant).where(MappedVariant.variant_id.in_(existing_variants)))
            db.execute(delete(Variant).where(Variant.id.in_(existing_variants)))
            logging_context["deleted_variants"] = score_set.num_variants
            score_set.num_variants = 0

            logger.info(msg="Deleted existing variants from score set.", extra=logging_context)

            db.commit()
            db.refresh(score_set)

        variants_data = create_variants_data(validated_scores, validated_counts, None)
        create_variants(db, score_set, variants_data)

    # Validation errors arise from problematic user data. These should be inserted into the database so failures can
    # be persisted to them.
    except ValidationError as e:
        db.rollback()
        score_set.processing_state = ProcessingState.failed
        score_set.processing_errors = {"exception": str(e), "detail": e.triggering_exceptions}
        score_set.mapping_state = MappingState.not_attempted

        logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
        logging_context["processing_state"] = score_set.processing_state.name
        logging_context["mapping_state"] = score_set.mapping_state.name
        logging_context["created_variants"] = 0
        logger.warning(msg="Encountered a validation error while processing variants.", extra=logging_context)

        return {"success": False}

    # NOTE: Since these are likely to be internal errors, it makes less sense to add them to the DB and surface them to the end user.
    # Catch all non-system exiting exceptions.
    except Exception as e:
        db.rollback()
        score_set.processing_state = ProcessingState.failed
        score_set.processing_errors = {"exception": str(e), "detail": []}
        score_set.mapping_state = MappingState.not_attempted

        logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
        logging_context["processing_state"] = score_set.processing_state.name
        logging_context["mapping_state"] = score_set.mapping_state.name
        logging_context["created_variants"] = 0
        logger.warning(msg="Encountered an internal exception while processing variants.", extra=logging_context)

        send_slack_message(err=e)
        return {"success": False}

    # Catch all other exceptions. The exceptions caught here were intented to be system exiting.
    except BaseException as e:
        db.rollback()
        score_set.processing_state = ProcessingState.failed
        score_set.mapping_state = MappingState.not_attempted
        db.commit()

        logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
        logging_context["processing_state"] = score_set.processing_state.name
        logging_context["mapping_state"] = score_set.mapping_state.name
        logging_context["created_variants"] = 0
        logger.error(
            msg="Encountered an unhandled exception while creating variants for score set.", extra=logging_context
        )

        # Don't raise BaseExceptions so we may emit canonical logs (TODO: Perhaps they are so problematic we want to raise them anyway).
        return {"success": False}

    else:
        score_set.processing_state = ProcessingState.success
        score_set.processing_errors = null()

        logging_context["created_variants"] = score_set.num_variants
        logging_context["processing_state"] = score_set.processing_state.name
        logger.info(msg="Finished creating variants in score set.", extra=logging_context)

        await redis.lpush(MAPPING_QUEUE_NAME, score_set.id)  # type: ignore
        await redis.enqueue_job("variant_mapper_manager", correlation_id, updater_id)
        score_set.mapping_state = MappingState.queued
    finally:
        db.add(score_set)
        db.commit()
        db.refresh(score_set)
        logger.info(msg="Committed new variants to score set.", extra=logging_context)

    ctx["state"][ctx["job_id"]] = logging_context.copy()
    return {"success": True}


async def map_variants_for_score_set(
    ctx: dict, correlation_id: str, score_set_id: int, updater_id: int, attempt: int = 1
) -> dict:
    async with mapping_in_execution(redis=ctx["redis"], job_id=ctx["job_id"]):
        logging_context = {}
        score_set = None
        try:
            db: Session = ctx["db"]
            redis: ArqRedis = ctx["redis"]
            score_set = db.scalars(select(ScoreSet).where(ScoreSet.id == score_set_id)).one()

            logging_context = setup_job_state(ctx, updater_id, score_set.urn, correlation_id)
            logging_context["attempt"] = attempt
            logger.info(msg="Started variant mapping", extra=logging_context)

            score_set.mapping_state = MappingState.processing
            score_set.mapping_errors = null()
            db.add(score_set)
            db.commit()

            mapping_urn = score_set.urn
            assert mapping_urn, "A valid URN is needed to map this score set."

            logging_context["current_mapping_resource"] = mapping_urn
            logging_context["mapping_state"] = score_set.mapping_state
            logger.debug(msg="Fetched score set metadata for mapping job.", extra=logging_context)

            # Do not block Worker event loop during mapping, see: https://arq-docs.helpmanual.io/#synchronous-jobs.
            vrs = vrs_mapper()
            blocking = functools.partial(vrs.map_score_set, mapping_urn)
            loop = asyncio.get_running_loop()

        except Exception as e:
            send_slack_message(e)
            logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
            logger.error(
                msg="Variant mapper encountered an unexpected error during setup. This job will not be retried.",
                extra=logging_context,
            )

            db.rollback()
            if score_set:
                score_set.mapping_state = MappingState.failed
                score_set.mapping_errors = {"error_message": "Encountered an internal server error during mapping"}
                db.add(score_set)
            db.commit()

            return {"success": False, "retried": False}

        mapping_results = None
        try:
            mapping_results = await loop.run_in_executor(ctx["pool"], blocking)
            logger.debug(msg="Done mapping variants.", extra=logging_context)

        except Exception as e:
            db.rollback()
            score_set.mapping_errors = {
                "error_message": f"Encountered an internal server error during mapping. Mapping will be automatically retried up to 5 times for this score set (attempt {attempt}/5)."
            }
            db.add(score_set)
            db.commit()

            send_slack_message(e)
            logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
            logger.warning(
                msg="Variant mapper encountered an unexpected error while mapping variants. This job will be retried.",
                extra=logging_context,
            )

            new_job_id = None
            max_retries_exceeded = None
            try:
                await redis.lpush(MAPPING_QUEUE_NAME, score_set.id)  # type: ignore
                new_job_id, max_retries_exceeded, backoff_time = await enqueue_job_with_backoff(
                    redis, "variant_mapper_manager", attempt, correlation_id, updater_id
                )
                # If we fail to enqueue a mapping manager for this score set, evict it from the queue.
                if new_job_id is None:
                    await redis.lpop(MAPPING_QUEUE_NAME, score_set.id)  # type: ignore

                logging_context["backoff_limit_exceeded"] = max_retries_exceeded
                logging_context["backoff_deferred_in_seconds"] = backoff_time
                logging_context["backoff_job_id"] = new_job_id

            except Exception as backoff_e:
                score_set.mapping_state = MappingState.failed
                score_set.mapping_errors = {"error_message": "Encountered an internal server error during mapping"}
                db.add(score_set)
                db.commit()
                send_slack_message(backoff_e)
                logging_context = {**logging_context, **format_raised_exception_info_as_dict(backoff_e)}
                logger.critical(
                    msg="While attempting to re-enqueue a mapping job that exited in error, another exception was encountered. This score set will not be mapped.",
                    extra=logging_context,
                )
            else:
                if new_job_id and not max_retries_exceeded:
                    score_set.mapping_state = MappingState.queued
                    db.add(score_set)
                    db.commit()
                    logger.info(
                        msg="After encountering an error while mapping variants, another mapping job was queued.",
                        extra=logging_context,
                    )
                elif new_job_id is None and not max_retries_exceeded:
                    score_set.mapping_state = MappingState.failed
                    score_set.mapping_errors = {"error_message": "Encountered an internal server error during mapping"}
                    db.add(score_set)
                    db.commit()
                    logger.error(
                        msg="After encountering an error while mapping variants, another mapping job was unable to be queued. This score set will not be mapped.",
                        extra=logging_context,
                    )
                else:
                    score_set.mapping_state = MappingState.failed
                    score_set.mapping_errors = {"error_message": "Encountered an internal server error during mapping"}
                    db.add(score_set)
                    db.commit()
                    logger.error(
                        msg="After encountering an error while mapping variants, the maximum retries for this job were exceeded. This score set will not be mapped.",
                        extra=logging_context,
                    )
            finally:
                return {"success": False, "retried": (not max_retries_exceeded and new_job_id is not None)}

        try:
            if mapping_results:
                mapped_scores = mapping_results.get("mapped_scores")
                if not mapped_scores:
                    # if there are no mapped scores, the score set failed to map.
                    score_set.mapping_state = MappingState.failed
                    score_set.mapping_errors = {"error_message": mapping_results.get("error_message")}
                else:
                    # TODO(VariantEffect/dcd-mapping2#2) after adding multi target mapping support:
                    # this assumes single-target mapping, will need to be changed to support multi-target mapping
                    # just in case there are multiple target genes in the db for a score set (this point shouldn't be reached
                    # while we only support single-target mapping), match up the target sequence with the one in the computed genomic reference sequence.
                    # TODO(VariantEffect/dcd-mapping2#3) after adding accession-based score set mapping support:
                    # this also assumes that the score set is based on a target sequence, not a target accession

                    computed_genomic_ref = mapping_results.get("computed_genomic_reference_sequence")
                    mapped_genomic_ref = mapping_results.get("mapped_genomic_reference_sequence")
                    computed_protein_ref = mapping_results.get("computed_protein_reference_sequence")
                    mapped_protein_ref = mapping_results.get("mapped_protein_reference_sequence")

                    if computed_genomic_ref:
                        target_sequence = computed_genomic_ref["sequence"]  # noqa: F841
                    elif computed_protein_ref:
                        target_sequence = computed_protein_ref["sequence"]  # noqa: F841
                    else:
                        raise NonexistentMappingReferenceError()

                    # TODO(VariantEffect/dcd_mapping2#2): Handle variant mappings for score sets with more than 1 target.
                    target_gene = score_set.target_genes[0]

                    excluded_pre_mapped_keys = {"sequence"}
                    if computed_genomic_ref and mapped_genomic_ref:
                        pre_mapped_metadata = computed_genomic_ref
                        target_gene.pre_mapped_metadata = cast(
                            {
                                "genomic": {
                                    k: pre_mapped_metadata[k]
                                    for k in set(list(pre_mapped_metadata.keys())) - excluded_pre_mapped_keys
                                }
                            },
                            JSONB,
                        )
                        target_gene.post_mapped_metadata = cast({"genomic": mapped_genomic_ref}, JSONB)
                    elif computed_protein_ref and mapped_protein_ref:
                        pre_mapped_metadata = computed_protein_ref
                        target_gene.pre_mapped_metadata = cast(
                            {
                                "protein": {
                                    k: pre_mapped_metadata[k]
                                    for k in set(list(pre_mapped_metadata.keys())) - excluded_pre_mapped_keys
                                }
                            },
                            JSONB,
                        )
                        target_gene.post_mapped_metadata = cast({"protein": mapped_protein_ref}, JSONB)
                    else:
                        raise NonexistentMappingReferenceError()

                    total_variants = 0
                    successful_mapped_variants = 0
                    for mapped_score in mapped_scores:
                        total_variants += 1
                        variant_urn = mapped_score.get("mavedb_id")
                        variant = db.scalars(select(Variant).where(Variant.urn == variant_urn)).one()

                        # there should only be one current mapped variant per variant id, so update old mapped variant to current = false
                        existing_mapped_variant = (
                            db.query(MappedVariant)
                            .filter(MappedVariant.variant_id == variant.id, MappedVariant.current.is_(True))
                            .one_or_none()
                        )

                        if existing_mapped_variant:
                            existing_mapped_variant.current = False
                            db.add(existing_mapped_variant)

                        if mapped_score.get("pre_mapped") and mapped_score.get("post_mapped"):
                            successful_mapped_variants += 1

                        mapped_variant = MappedVariant(
                            pre_mapped=mapped_score.get("pre_mapped", null()),
                            post_mapped=mapped_score.get("post_mapped", null()),
                            variant_id=variant.id,
                            modification_date=date.today(),
                            mapped_date=mapping_results["mapped_date_utc"],
                            vrs_version=mapped_score.get("vrs_version", null()),
                            mapping_api_version=mapping_results["dcd_mapping_version"],
                            error_message=mapped_score.get("error_message", null()),
                            current=True,
                        )
                        db.add(mapped_variant)

                    if successful_mapped_variants == 0:
                        score_set.mapping_state = MappingState.failed
                        score_set.mapping_errors = {"error_message": "All variants failed to map"}
                    elif successful_mapped_variants < total_variants:
                        score_set.mapping_state = MappingState.incomplete
                    else:
                        score_set.mapping_state = MappingState.complete

                    logging_context["mapped_variants_inserted_db"] = len(mapped_scores)
                    logging_context["variants_successfully_mapped"] = successful_mapped_variants
                    logging_context["mapping_state"] = score_set.mapping_state.name
                    logging_context["mapping_errors"] = score_set.mapping_errors
                    logger.info(msg="Inserted mapped variants into db.", extra=logging_context)

            else:
                raise NonexistentMappingResultsError()

            db.add(score_set)
            db.commit()

        except Exception as e:
            db.rollback()
            score_set.mapping_errors = {
                "error_message": f"Encountered an unexpected error while parsing mapped variants. Mapping will be automatically retried up to 5 times for this score set (attempt {attempt}/5)."
            }
            db.add(score_set)
            db.commit()

            send_slack_message(e)
            logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
            logger.warning(
                msg="An unexpected error occurred during variant mapping. This job will be attempted again.",
                extra=logging_context,
            )

            new_job_id = None
            max_retries_exceeded = None
            try:
                await redis.lpush(MAPPING_QUEUE_NAME, score_set.id)  # type: ignore
                new_job_id, max_retries_exceeded, backoff_time = await enqueue_job_with_backoff(
                    redis, "variant_mapper_manager", attempt, correlation_id, updater_id
                )
                # If we fail to enqueue a mapping manager for this score set, evict it from the queue.
                if new_job_id is None:
                    await redis.lpop(MAPPING_QUEUE_NAME, score_set.id)  # type: ignore

                logging_context["backoff_limit_exceeded"] = max_retries_exceeded
                logging_context["backoff_deferred_in_seconds"] = backoff_time
                logging_context["backoff_job_id"] = new_job_id

            except Exception as backoff_e:
                score_set.mapping_state = MappingState.failed
                score_set.mapping_errors = {"error_message": "Encountered an internal server error during mapping"}
                send_slack_message(backoff_e)
                logging_context = {**logging_context, **format_raised_exception_info_as_dict(backoff_e)}
                logger.critical(
                    msg="While attempting to re-enqueue a mapping job that exited in error, another exception was encountered. This score set will not be mapped.",
                    extra=logging_context,
                )
            else:
                if new_job_id and not max_retries_exceeded:
                    score_set.mapping_state = MappingState.queued
                    logger.info(
                        msg="After encountering an error while parsing mapped variants, another mapping job was queued.",
                        extra=logging_context,
                    )
                elif new_job_id is None and not max_retries_exceeded:
                    score_set.mapping_state = MappingState.failed
                    score_set.mapping_errors = {"error_message": "Encountered an internal server error during mapping"}
                    logger.error(
                        msg="After encountering an error while parsing mapped variants, another mapping job was unable to be queued. This score set will not be mapped.",
                        extra=logging_context,
                    )
                else:
                    score_set.mapping_state = MappingState.failed
                    score_set.mapping_errors = {"error_message": "Encountered an internal server error during mapping"}
                    logger.error(
                        msg="After encountering an error while parsing mapped variants, the maximum retries for this job were exceeded. This score set will not be mapped.",
                        extra=logging_context,
                    )
            finally:
                db.add(score_set)
                db.commit()
                return {"success": False, "retried": (not max_retries_exceeded and new_job_id is not None)}

    ctx["state"][ctx["job_id"]] = logging_context.copy()
    return {"success": True, "retried": False}


async def variant_mapper_manager(ctx: dict, correlation_id: str, updater_id: int, attempt: int = 1) -> dict:
    logging_context = {}
    mapping_job_id = None
    mapping_job_status = None
    queued_score_set = None
    try:
        redis: ArqRedis = ctx["redis"]
        db: Session = ctx["db"]

        logging_context = setup_job_state(ctx, updater_id, None, correlation_id)
        logging_context["attempt"] = attempt
        logger.debug(msg="Variant mapping manager began execution", extra=logging_context)

        queue_length = await redis.llen(MAPPING_QUEUE_NAME)  # type: ignore
        queued_id = await redis.rpop(MAPPING_QUEUE_NAME)  # type: ignore
        logging_context["variant_mapping_queue_length"] = queue_length

        # Setup the job id cache if it does not already exist.
        if not await redis.exists(MAPPING_CURRENT_ID_NAME):
            await redis.set(MAPPING_CURRENT_ID_NAME, "")

        if not queued_id:
            logger.debug(msg="No mapping jobs exist in the queue.", extra=logging_context)
            return {"success": True, "enqueued_job": None}
        else:
            queued_id = queued_id.decode("utf-8")
            queued_score_set = db.scalars(select(ScoreSet).where(ScoreSet.id == queued_id)).one()

            logging_context["upcoming_mapping_resource"] = queued_score_set.urn
            logger.debug(msg="Found mapping job(s) still in queue.", extra=logging_context)

        mapping_job_id = await redis.get(MAPPING_CURRENT_ID_NAME)
        if mapping_job_id:
            mapping_job_id = mapping_job_id.decode("utf-8")
            mapping_job_status = (await Job(job_id=mapping_job_id, redis=redis).status()).value

        logging_context["existing_mapping_job_status"] = mapping_job_status
        logging_context["existing_mapping_job_id"] = mapping_job_id

    except Exception as e:
        send_slack_message(e)

        # Attempt to remove this item from the mapping queue.
        try:
            await redis.lrem(MAPPING_QUEUE_NAME, 1, queued_id)  # type: ignore
            logger.warning(msg="Removed un-queueable score set from the queue.", extra=logging_context)
        except Exception:
            pass

        logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
        logger.error(msg="Variant mapper manager encountered an unexpected error during setup.", extra=logging_context)

        return {"success": False, "enqueued_job": None}

    new_job = None
    new_job_id = None
    try:
        if not mapping_job_id or mapping_job_status in (JobStatus.not_found, JobStatus.complete):
            logger.debug(msg="No mapping jobs are running, queuing a new one.", extra=logging_context)

            new_job = await redis.enqueue_job(
                "map_variants_for_score_set", correlation_id, queued_score_set.id, updater_id, attempt
            )

        if new_job:
            new_job_id = new_job.job_id

            logging_context["new_mapping_job_id"] = new_job_id
            logger.info(msg="Queued a new mapping job.", extra=logging_context)

            return {"success": True, "enqueued_job": new_job_id}

        logger.info(
            msg="A mapping job is already running, or a new job was unable to be enqueued. Deferring mapping by 5 minutes.",
            extra=logging_context,
        )

        new_job = await redis.enqueue_job(
            "variant_mapper_manager",
            correlation_id,
            updater_id,
            attempt,
            _defer_by=timedelta(minutes=5),
        )

        if new_job:
            # Ensure this score set remains in the front of the queue.
            queued_id = await redis.rpush(MAPPING_QUEUE_NAME, queued_score_set.id)  # type: ignore
            new_job_id = new_job.job_id

            logging_context["new_mapping_manager_job_id"] = new_job_id
            logger.info(msg="Deferred a new mapping manager job.", extra=logging_context)

            # Our persistent Redis queue and ARQ's execution rules ensure that even if the worker is stopped and not restarted
            # before the deferred time, these deferred jobs will still run once able.
            return {"success": True, "enqueued_job": new_job_id}

        raise MappingEnqueueError()

    except Exception as e:
        send_slack_message(e)
        logging_context = {**logging_context, **format_raised_exception_info_as_dict(e)}
        logger.error(
            msg="Variant mapper manager encountered an unexpected error while enqueing a mapping job. This job will not be retried.",
            extra=logging_context,
        )

        db.rollback()

        # We shouldn't rely on the passed score set id matching the score set we are operating upon.
        if not queued_score_set:
            return {"success": False, "enqueued_job": new_job_id}

        # Attempt to remove this item from the mapping queue.
        try:
            await redis.lrem(MAPPING_QUEUE_NAME, 1, queued_id)  # type: ignore
            logger.warning(msg="Removed un-queueable score set from the queue.", extra=logging_context)
        except Exception:
            pass

        score_set_exc = db.scalars(select(ScoreSet).where(ScoreSet.id == queued_score_set.id)).one_or_none()
        if score_set_exc:
            score_set_exc.mapping_state = MappingState.failed
            score_set_exc.mapping_errors = "Unable to queue a new mapping job or defer score set mapping."
            db.add(score_set_exc)
        db.commit()

        return {"success": False, "enqueued_job": new_job_id}



"""
Environment setup for scripts.
"""

import enum
import logging
import click
from functools import wraps


from sqlalchemy.orm import configure_mappers

from mavedb import deps
from mavedb.models import *  # noqa: F403


logger = logging.getLogger(__name__)


@enum.unique
class DatabaseSessionAction(enum.Enum):
    """
    Enum representing the database session transaction action selected for a
    command decorated by :py:func:`.with_database_session`.

    You will not need to use this class unless you provide ``pass_action =
    True`` to :py:func:`.with_database_session`.
    """

    DRY_RUN = "rollback"
    PROMPT = "prompt"
    COMMIT = "commit"


@click.group()
def script_environment():
    """
    Set up the environment for a script that may be run from the command line and does not necessarily depend on the
    FastAPI framework.

    Features:
    - Configures logging for the script.
    - Loads the SQLAlchemy data model.
    """

    logging.basicConfig()

    # Un-comment this line to log all database queries:
    logging.getLogger("__main__").setLevel(logging.INFO)
    # logging.getLogger("sqlalchemy.engine").setLevel(logging.INFO)

    # Scan all our model classes and create backref attributes. Otherwise, these attributes only get added to classes once
    # an instance of the related class has been created.
    configure_mappers()


def with_database_session(command=None, *, pass_action: bool = False):
    """
    Decorator to provide database session and error handling for a *command*.

    The *command* callable must be a :py:class:`click.Command` instance.

    The decorated *command* is called with a ``db`` keyword argument to provide
    a :class:`~id3c.db.session.DatabaseSession` object.  The call happens
    within an exception handler that commits or rollsback the database
    transaction, possibly interactively.  Three new options are added to the
    *command* (``--dry-run``, ``--prompt``, and ``--commit``) to control this
    behaviour.

    >>> @click.command
    ... @with_database_session
    ... def cmd(db: DatabaseSession):
    ...     pass

    If the optional, keyword-only argument *pass_action* is ``True``, then the
    :py:class:`.DatabaseSessionAction` selected by the CLI options above is
    passed as an additional ``action`` argument to the decorated *command*.

    >>> @click.command
    ... @with_database_session(pass_action = True)
    ... def cmd(db: DatabaseSession, action: DatabaseSessionAction):
    ...     pass

    One example where this is useful is when the *command* accesses
    non-database resources and wants to extend dry run mode to them as well.
    """

    def decorator(command):
        @click.option(
            "--dry-run",
            "action",
            help="Only go through the motions of changing the database (default)",
            flag_value=DatabaseSessionAction("rollback"),
            type=DatabaseSessionAction,
            default=True,
        )
        @click.option(
            "--prompt",
            "action",
            help="Ask if changes to the database should be saved",
            flag_value=DatabaseSessionAction("prompt"),
            type=DatabaseSessionAction,
        )
        @click.option(
            "--commit",
            "action",
            help="Save changes to the database",
            flag_value=DatabaseSessionAction("commit"),
            type=DatabaseSessionAction,
        )
        @wraps(command)
        def decorated(*args, action, **kwargs):
            db = next(deps.get_db())

            kwargs["db"] = db

            if pass_action:
                kwargs["action"] = action

            processed_without_error = None

            try:
                command(*args, **kwargs)

            except Exception as error:
                processed_without_error = False

                logger.error(f"Aborting with error: {error}")
                raise error from None

            else:
                processed_without_error = True

            finally:
                if action is DatabaseSessionAction.PROMPT:
                    ask_to_commit = (
                        "Commit all changes?"
                        if processed_without_error
                        else "Commit successfully processed records up to this point?"
                    )

                    commit = click.confirm(ask_to_commit)
                else:
                    commit = action is DatabaseSessionAction.COMMIT

                if commit:
                    logger.info(
                        "Committing all changes"
                        if processed_without_error
                        else "Committing successfully processed records up to this point"
                    )
                    db.commit()

                else:
                    logger.info("Rolling back all changes; the database will not be modified")
                    db.rollback()

        return decorated

    return decorator(command) if command else decorator



from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import UserData
from mavedb.lib.authorization import require_current_user
from mavedb.lib.target_genes import (
    search_target_genes as _search_target_genes,
)
from mavedb.models.target_gene import TargetGene
from mavedb.view_models import target_gene
from mavedb.view_models.search import TextSearch

router = APIRouter(prefix="/api/v1", tags=["target-genes"], responses={404: {"description": "Not found"}})


@router.post("/me/target-genes/search", status_code=200, response_model=List[target_gene.TargetGene])
def search_my_target_genes(
    search: TextSearch,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user)
) -> Any:
    """
    Search my target genes.
    """

    return _search_target_genes(db, user_data.user, search, 50)


@router.get("/target-genes", status_code=200, response_model=List[target_gene.TargetGene], responses={404: {}})
def list_target_genes(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List target genes.
    """
    items = db.query(TargetGene).order_by(TargetGene.name).all()
    return items


@router.get("/target-genes/names", status_code=200, response_model=List[str], responses={404: {}})
def list_target_gene_names(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List distinct target gene names, in alphabetical order.
    """

    items = db.query(TargetGene).all()
    names = map(lambda item: item.name, items)
    return sorted(list(set(names)))


@router.get("/target-genes/categories", status_code=200, response_model=List[str], responses={404: {}})
def list_target_gene_categories(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List distinct target genes categories, in alphabetical order.
    """

    items = db.query(TargetGene).all()
    categories = map(lambda item: item.category, items)
    return sorted(list(set(categories)))


@router.get("/target-genes/{item_id}", status_code=200, response_model=target_gene.TargetGene, responses={404: {}})
def fetch_target_gene(
    *,
    item_id: int,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single target gene by ID.
    """
    item = db.query(TargetGene).filter(TargetGene.id == item_id).first()
    if not item:
        raise HTTPException(status_code=404, detail=f"TargetGene with ID {item_id} not found")
    return item


@router.post("/target-genes/search", status_code=200, response_model=List[target_gene.TargetGene])
def search_target_genes(
    search: TextSearch,
    db: Session = Depends(deps.get_db)
) -> Any:
    """
    Search target genes.
    """

    return _search_target_genes(db, None, search, 50)



from typing import Any

import httpx
from fastapi.exceptions import HTTPException
from sqlalchemy.orm import Session

from mavedb.models.taxonomy import Taxonomy
from mavedb.view_models.taxonomy import TaxonomyCreate


async def find_or_create_taxonomy(db: Session, taxonomy: TaxonomyCreate):
    """
    Find an existing taxonomy ID record with the specified tax_id int, or create a new one.

    :param db: An active database session
    :param taxonomy: A TaxonomyCreate object containing the taxonomy details to search for or create.
    :return: An existing Taxonomy containing the specified taxonomy ID, or a new, unsaved Taxonomy
    tax_id: A valid taxonomy ID from NCBI
    """
    taxonomy_record = db.query(Taxonomy).filter(Taxonomy.tax_id == taxonomy.tax_id).one_or_none()
    if not taxonomy_record:
        taxonomy_record = await search_NCBI_taxonomy(db, str(taxonomy.tax_id))
    return taxonomy_record


async def search_NCBI_taxonomy(db: Session, search: str) -> Any:
    """
    Check whether this taxonomy exists in NCBI website.
    If it exists, saving it in local database and return it to users.
    """
    if search and len(search.strip()) > 0:
        if search.isnumeric() is False:
            search_text = search.strip().lower()
        else:
            search_text = str(search)

        url = f"https://api.ncbi.nlm.nih.gov/datasets/v2alpha/taxonomy/taxon/{search_text}"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            if response.status_code == 200:
                data = response.json()

                # response status code is 200 even no match data in NCBI.
                if "errors" in data.get("taxonomy_nodes", [{}])[0]:
                    return None

                # Process the retrieved data as needed
                ncbi_taxonomy = data["taxonomy_nodes"][0]["taxonomy"]
                ncbi_taxonomy.setdefault("organism_name", "NULL")
                ncbi_taxonomy.setdefault("common_name", "NULL")
                ncbi_taxonomy.setdefault("rank", "NULL")
                ncbi_taxonomy.setdefault("has_described_species_name", False)
                taxonomy_record = Taxonomy(
                    tax_id=ncbi_taxonomy["tax_id"],
                    organism_name=ncbi_taxonomy["organism_name"],
                    common_name=ncbi_taxonomy["common_name"],
                    rank=ncbi_taxonomy["rank"],
                    has_described_species_name=ncbi_taxonomy["has_described_species_name"],
                    url="https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?mode=info&id="
                    + str(ncbi_taxonomy["tax_id"]),
                    article_reference="NCBI:txid" + str(ncbi_taxonomy["tax_id"]),
                )
                db.add(taxonomy_record)
                db.commit()
                db.refresh(taxonomy_record)
            else:
                raise HTTPException(status_code=404, detail=f"Taxonomy with search {search_text} not found in NCBI")
    else:
        raise HTTPException(status_code=404, detail="Please enter valid searching words")

    return taxonomy_record



# See https://pydantic-docs.helpmanual.io/usage/postponed_annotations/#self-referencing-models
from __future__ import annotations

from datetime import date
from typing import Any, Collection, Dict, Optional, Sequence

from humps import camelize
from pydantic import root_validator

from mavedb.lib.validation import urn_re
from mavedb.lib.validation.constants.score_set import default_ranges
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.utilities import inf_or_float, is_null
from mavedb.models.enums.mapping_state import MappingState
from mavedb.models.enums.processing_state import ProcessingState
from mavedb.view_models import PublicationIdentifiersGetter, record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models.contributor import Contributor, ContributorCreate
from mavedb.view_models.doi_identifier import (
    DoiIdentifier,
    DoiIdentifierCreate,
    SavedDoiIdentifier,
)
from mavedb.view_models.license import ShortLicense
from mavedb.view_models.publication_identifier import (
    PublicationIdentifier,
    PublicationIdentifierCreate,
    SavedPublicationIdentifier,
)
from mavedb.view_models.target_gene import (
    SavedTargetGene,
    ShortTargetGene,
    TargetGene,
    TargetGeneCreate,
)
from mavedb.view_models.user import SavedUser, User
from mavedb.view_models.variant import VariantInDbBase


class ExternalLink(BaseModel):
    url: Optional[str]


class ScoreRange(BaseModel):
    label: str
    description: Optional[str]
    classification: str
    # Purposefully vague type hint because of some odd JSON Schema generation behavior.
    # Typing this as tuple[Union[float, None], Union[float, None]] will generate an invalid
    # jsonschema, and fail all tests that access the schema. This may be fixed in pydantic v2,
    # but it's unclear. Even just typing it as Tuple[Any, Any] will generate an invalid schema!
    range: list[Any]  # really: tuple[Union[float, None], Union[float, None]]

    @validator("classification")
    def range_classification_value_is_accepted(cls, field_value: str):
        classification = field_value.strip().lower()
        if classification not in default_ranges:
            raise ValidationError(
                f"Unexpected classification value(s): {classification}. Permitted values: {default_ranges}"
            )

        return classification

    @validator("range")
    def ranges_are_not_backwards(cls, field_value: tuple[Any]):
        if len(field_value) != 2:
            raise ValidationError("Only a lower and upper bound are allowed.")

        field_value[0] = inf_or_float(field_value[0], True) if field_value[0] is not None else None
        field_value[1] = inf_or_float(field_value[1], False) if field_value[1] is not None else None

        if inf_or_float(field_value[0], True) > inf_or_float(field_value[1], False):
            raise ValidationError("The lower bound of the score range may not be larger than the upper bound.")
        elif inf_or_float(field_value[0], True) == inf_or_float(field_value[1], False):
            raise ValidationError("The lower and upper bound of the score range may not be the same.")

        return field_value


class ScoreRanges(BaseModel):
    wt_score: float
    ranges: list[ScoreRange]  # type: ignore


class ScoreSetGetter(PublicationIdentifiersGetter):
    def get(self, key: Any, default: Any = ...) -> Any:
        if key == "meta_analyzes_score_set_urns":
            meta_analyzes_score_sets = getattr(self._obj, "meta_analyzes_score_sets") or []
            return sorted([score_set.urn for score_set in meta_analyzes_score_sets])
        elif key == "meta_analyzed_by_score_set_urns":
            meta_analyzed_by_score_sets = getattr(self._obj, "meta_analyzed_by_score_sets") or []
            return sorted([score_set.urn for score_set in meta_analyzed_by_score_sets])
        else:
            return super().get(key, default)


class ScoreSetBase(BaseModel):
    """Base class for score set view models."""

    title: str
    method_text: str
    abstract_text: str
    short_description: str
    extra_metadata: Optional[dict]
    data_usage_policy: Optional[str]


class ScoreSetModify(ScoreSetBase):
    contributors: Optional[list[ContributorCreate]]
    primary_publication_identifiers: Optional[list[PublicationIdentifierCreate]]
    secondary_publication_identifiers: Optional[list[PublicationIdentifierCreate]]
    doi_identifiers: Optional[list[DoiIdentifierCreate]]
    target_genes: list[TargetGeneCreate]
    score_ranges: Optional[ScoreRanges]

    @validator("title", "short_description", "abstract_text", "method_text")
    def validate_field_is_non_empty(cls, v):
        if is_null(v) or not isinstance(v, str):
            raise ValidationError("This field is required and cannot be empty.")
        return v.strip()

    @validator("primary_publication_identifiers")
    def max_one_primary_publication_identifier(cls, v):
        if isinstance(v, list):
            if len(v) > 1:
                raise ValidationError("multiple primary publication identifiers are not allowed")
        return v

    # Validate nested label field within target sequence if there are multiple target genes
    @validator("target_genes")
    def targets_need_labels_when_multiple_targets_exist(cls, field_value, values):
        if len(field_value) > 1:
            for idx, target in enumerate(field_value):
                if target.target_sequence and target.target_sequence.label is None:
                    raise ValidationError(
                        "Target sequence labels cannot be empty when multiple targets are defined.",
                        custom_loc=[
                            "body",
                            "targetGene",
                            idx,
                            "targetSequence",
                            "label",
                        ],
                    )

        return field_value

    # Validate nested label fields are not identical
    @validator("target_genes")
    def target_labels_are_unique(cls, field_value, values):
        # Labels are only used on target sequence instances.
        if len(field_value) > 1 and all([target.target_sequence is not None for target in field_value]):
            # Labels have already been sanitized by the TargetSequence validator.
            labels = [target.target_sequence.label for target in field_value]
            dup_indices = [idx for idx, item in enumerate(labels) if item in labels[:idx]]
            if dup_indices:
                # TODO: surface the error for the each duplicated index. the way these pydantic validators are
                # implemented would require additional customization to surface each duplicate, so surfacing
                # just one for now seems fine.
                raise ValidationError(
                    "Target sequence labels cannot be duplicated.",
                    custom_loc=[
                        "body",
                        "targetGene",
                        dup_indices[-1],
                        "targetSequence",
                        "label",
                    ],
                )

        return field_value

    # Validate that this score set contains at least one target attached to it
    @validator("target_genes")
    def at_least_one_target_gene_exists(cls, field_value, values):
        if len(field_value) < 1:
            raise ValidationError("Score sets should define at least one target.")

        return field_value

    @validator("score_ranges")
    def score_range_labels_must_be_unique(cls, field_value: Optional[ScoreRanges]):
        if field_value is None:
            return None

        existing_labels = []
        for i, range_model in enumerate(field_value.ranges):
            range_model.label = range_model.label.strip()

            if range_model.label in existing_labels:
                raise ValidationError(
                    f"Detected repeated label: `{range_model.label}`. Range labels must be unique.",
                    custom_loc=["body", "scoreRanges", "ranges", i, "label"],
                )

            existing_labels.append(range_model.label)

        return field_value

    @validator("score_ranges")
    def ranges_contain_normal_and_abnormal(cls, field_value: Optional[ScoreRanges]):
        if field_value is None:
            return None

        ranges = set([range_model.classification for range_model in field_value.ranges])
        if not set(default_ranges).issubset(ranges):
            raise ValidationError(
                "Both `normal` and `abnormal` ranges must be provided.",
                # Raise this error inside the first classification provided by the model.
                custom_loc=["body", "scoreRanges", "ranges", 0, "classification"],
            )

        return field_value

    @validator("score_ranges")
    def ranges_do_not_overlap(cls, field_value: Optional[ScoreRanges]):
        def test_overlap(tp1, tp2) -> bool:
            # Always check the tuple with the lowest lower bound. If we do not check
            # overlaps in this manner, checking the overlap of (0,1) and (1,2) will
            # yield different results depending on the ordering of tuples.
            if min(inf_or_float(tp1[0], True), inf_or_float(tp2[0], True)) == inf_or_float(tp1[0], True):
                tp_with_min_value = tp1
                tp_with_non_min_value = tp2
            else:
                tp_with_min_value = tp2
                tp_with_non_min_value = tp1

            if inf_or_float(tp_with_min_value[1], False) > inf_or_float(
                tp_with_non_min_value[0], True
            ) and inf_or_float(tp_with_min_value[0], True) <= inf_or_float(tp_with_non_min_value[1], False):
                return True

            return False

        if field_value is None:
            return None

        for i, range_test in enumerate(field_value.ranges):
            for range_check in list(field_value.ranges)[i + 1 :]:
                if test_overlap(range_test.range, range_check.range):
                    raise ValidationError(
                        f"Score ranges may not overlap; `{range_test.label}` overlaps with `{range_check.label}`",
                        custom_loc=["body", "scoreRanges", "ranges", i, "range"],
                    )

        return field_value

    @validator("score_ranges")
    def wild_type_score_in_normal_range(cls, field_value: Optional[ScoreRanges]):
        if field_value is None:
            return None

        normal_ranges = [
            range_model.range for range_model in field_value.ranges if range_model.classification == "normal"
        ]
        for range in normal_ranges:
            if field_value.wt_score >= inf_or_float(range[0], lower=True) and field_value.wt_score < inf_or_float(
                range[1], lower=False
            ):
                return field_value

        raise ValidationError(
            f"The provided wild type score of {field_value.wt_score} is not within any of the provided normal ranges. This score should be within a normal range.",
            custom_loc=["body", "scoreRanges", "wtScore"],
        )


class ScoreSetCreate(ScoreSetModify):
    """View model for creating a new score set."""

    experiment_urn: Optional[str]
    license_id: int
    superseded_score_set_urn: Optional[str]
    meta_analyzes_score_set_urns: Optional[list[str]]

    @validator("superseded_score_set_urn")
    def validate_superseded_score_set_urn(cls, v):
        if v is None:
            pass
        else:
            if urn_re.MAVEDB_SCORE_SET_URN_RE.fullmatch(v) is None:
                if urn_re.MAVEDB_TMP_URN_RE.fullmatch(v) is None:
                    raise ValueError(f"'{v}' is not a valid score set URN")
                else:
                    raise ValueError("cannot supersede a private score set - please edit it instead")
        return v

    @validator("meta_analyzes_score_set_urns")
    def validate_score_set_urn(cls, v):
        if v is not None:
            for s in v:
                if (
                    urn_re.MAVEDB_SCORE_SET_URN_RE.fullmatch(s) is None
                    and urn_re.MAVEDB_TMP_URN_RE.fullmatch(s) is None
                ):
                    raise ValueError(f"'{s}' is not a valid score set URN")
        return v

    @validator("experiment_urn")
    def validate_experiment_urn(cls, v):
        if urn_re.MAVEDB_EXPERIMENT_URN_RE.fullmatch(v) is None and urn_re.MAVEDB_TMP_URN_RE.fullmatch(v) is None:
            raise ValueError(f"'{v}' is not a valid experiment URN")
        return v

    @root_validator
    def validate_experiment_urn_required_except_for_meta_analyses(cls, values):
        experiment_urn = values.get("experiment_urn")
        meta_analyzes_score_set_urns = values.get("meta_analyzes_score_set_urns")
        is_meta_analysis = not (meta_analyzes_score_set_urns is None or len(meta_analyzes_score_set_urns) == 0)
        if experiment_urn is None and not is_meta_analysis:
            raise ValidationError("experiment URN is required unless your score set is a meta-analysis")
        if experiment_urn is not None and is_meta_analysis:
            raise ValidationError("experiment URN should not be supplied when your score set is a meta-analysis")
        return values


class ScoreSetUpdate(ScoreSetModify):
    """View model for updating a score set."""

    license_id: Optional[int]


class ShortScoreSet(BaseModel):
    """
    Score set view model containing a smaller set of properties to return in list contexts.

    Notice that this is not derived from ScoreSetBase.
    """

    urn: str
    title: str
    short_description: str
    published_date: Optional[date]
    replaces_id: Optional[int]
    num_variants: int
    experiment: "Experiment"
    primary_publication_identifiers: list[SavedPublicationIdentifier]
    secondary_publication_identifiers: list[SavedPublicationIdentifier]
    license: ShortLicense
    creation_date: date
    modification_date: date
    target_genes: list[ShortTargetGene]
    private: bool
    record_type: str = None  # type: ignore

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True
        getter_dict = ScoreSetGetter


class ShorterScoreSet(BaseModel):
    urn: str
    record_type: str = None  # type: ignore

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True
        getter_dict = ScoreSetGetter


class SavedScoreSet(ScoreSetBase):
    """Base class for score set view models representing saved records."""

    record_type: str = None  # type: ignore
    urn: str
    num_variants: int
    license: ShortLicense
    superseded_score_set: Optional[ShorterScoreSet]
    superseding_score_set: Optional[ShorterScoreSet]
    meta_analyzes_score_set_urns: list[str]
    meta_analyzed_by_score_set_urns: list[str]
    doi_identifiers: Sequence[SavedDoiIdentifier]
    primary_publication_identifiers: Sequence[SavedPublicationIdentifier]
    secondary_publication_identifiers: Sequence[SavedPublicationIdentifier]
    published_date: Optional[date]
    creation_date: date
    modification_date: date
    created_by: Optional[SavedUser]
    modified_by: Optional[SavedUser]
    target_genes: Sequence[SavedTargetGene]
    dataset_columns: Dict
    external_links: Dict[str, ExternalLink]
    contributors: list[Contributor]
    score_ranges: Optional[ScoreRanges]

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True
        getter_dict = ScoreSetGetter

    # Association proxy objects return an untyped _AssocitionList object.
    # Recast it into something more generic.
    @validator("secondary_publication_identifiers", "primary_publication_identifiers", pre=True)
    def publication_identifiers_validator(cls, value) -> list[PublicationIdentifier]:
        assert isinstance(value, Collection), "Publication identifier lists must be a collection"
        return list(value)  # Re-cast into proper list-like type

    @validator("dataset_columns")
    def camelize_dataset_columns_keys(cls, value) -> dict:
        return camelize(value)


class ScoreSet(SavedScoreSet):
    """Score set view model containing most properties visible to non-admin users, but no variant data."""

    experiment: "Experiment"
    doi_identifiers: Sequence[DoiIdentifier]
    primary_publication_identifiers: Sequence[PublicationIdentifier]
    secondary_publication_identifiers: Sequence[PublicationIdentifier]
    created_by: Optional[User]
    modified_by: Optional[User]
    target_genes: Sequence[TargetGene]
    private: bool
    processing_state: Optional[ProcessingState]
    processing_errors: Optional[dict]
    mapping_state: Optional[MappingState]
    mapping_errors: Optional[dict]


class ScoreSetWithVariants(ScoreSet):
    """
    Score set view model containing a complete set of properties visible to non-admin users, for contexts where variants
    are requested.
    """

    variants: list[VariantInDbBase]


class AdminScoreSet(ScoreSet):
    """Score set view model containing properties to return to admin clients."""

    normalised: bool
    approved: bool


class ScoreSetPublicDump(SavedScoreSet):
    """Score set view model containing properties to include in a dump of all published data."""

    doi_identifiers: Sequence[DoiIdentifier]
    primary_publication_identifiers: Sequence[PublicationIdentifier]
    secondary_publication_identifiers: Sequence[PublicationIdentifier]
    created_by: Optional[User]
    modified_by: Optional[User]
    target_genes: Sequence[TargetGene]
    private: bool
    processing_state: Optional[ProcessingState]
    processing_errors: Optional[Dict]
    mapping_state: Optional[MappingState]
    mapping_errors: Optional[Dict]


# ruff: noqa: E402
from mavedb.view_models.experiment import Experiment

ShortScoreSet.update_forward_refs()
ScoreSet.update_forward_refs()



from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, backref, relationship

from mavedb.db.base import Base
from mavedb.models.controlled_keyword import ControlledKeyword

if TYPE_CHECKING:
    from mavedb.models.experiment import Experiment


class ExperimentControlledKeywordAssociation(Base):
    __tablename__ = "experiment_controlled_keywords"

    controlled_keyword_id = Column(Integer, ForeignKey("controlled_keywords.id"), nullable=False, primary_key=True)
    controlled_keyword: Mapped[ControlledKeyword] = relationship(
        "ControlledKeyword", backref=backref("experiment_controlled_keywords", uselist=True)
    )

    experiment_id = Column(Integer, ForeignKey("experiments.id"), nullable=False, primary_key=True)
    experiment: Mapped["Experiment"] = relationship(back_populates="keyword_objs")

    description = Column(String, nullable=True)



from typing import Optional

from pydantic import root_validator

from mavedb.lib.validation import keywords
from mavedb.view_models import keyword, record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel


class ExperimentControlledKeywordBase(BaseModel):
    """Base class for experiment and controlled keyword bridge table view models."""

    keyword: keyword.KeywordBase
    description: Optional[str]

    @root_validator(pre=True)
    def validate_fields(cls, values):
        validated_keyword = values.get("keyword")
        validated_description = values.get("description")

        # validated_keyword possible value: {'key': 'Delivery method', 'value': None}
        # Validate if keyword value is other, whether description is None.
        if validated_keyword and validated_keyword["value"]:
            keywords.validate_description(validated_keyword["value"], validated_keyword["key"], validated_description)
        return values


class ExperimentControlledKeywordCreate(ExperimentControlledKeywordBase):
    """View model for creating a new keyword."""

    keyword: keyword.KeywordCreate


class ExperimentControlledKeywordUpdate(ExperimentControlledKeywordBase):
    """View model for updating a keyword."""

    pass


class SavedExperimentControlledKeyword(ExperimentControlledKeywordBase):
    """Base class for keyword view models representing saved records."""

    record_type: str = None  # type: ignore

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True


class ExperimentControlledKeyword(SavedExperimentControlledKeyword):
    """Keyword view model for non-admin clients."""

    pass



from datetime import date
from typing import Any, Optional, Sequence

from pydantic import root_validator
from pydantic.utils import GetterDict

from mavedb.models.enums.target_category import TargetCategory
from mavedb.view_models import external_gene_identifier_offset, record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models.target_accession import SavedTargetAccession, TargetAccession, TargetAccessionCreate
from mavedb.view_models.target_sequence import (
    AdminTargetSequence,
    SavedTargetSequence,
    TargetSequence,
    TargetSequenceCreate,
)


class ExternalIdentifiersGetter(GetterDict):
    """
    Custom class used in transforming TargetGene SQLAlchemy model objects into Pydantic view model objects, with special
    handling of external identifiers for target genes.

    Pydantic uses GetterDict objects to access source objects as dictionaries, which can then be turned into Pydantic
    view model objects. Here we need to remap the underlying SQLAlchemy model's separate properties for Ensembl, RefSeq,
    and UniProt identifiers into an array of external identifiers.
    """

    def get(self, key: Any, default: Any = ...) -> Any:
        if key == "external_identifiers":
            ensembl_offset = getattr(self._obj, "ensembl_offset")
            refseq_offset = getattr(self._obj, "refseq_offset")
            uniprot_offset = getattr(self._obj, "uniprot_offset")
            return list(filter(lambda x: x is not None, [ensembl_offset, refseq_offset, uniprot_offset]))
        else:
            return super().get(key, default)


class TargetGeneBase(BaseModel):
    """Base class for target gene view models."""

    name: str
    category: TargetCategory
    external_identifiers: Sequence[external_gene_identifier_offset.ExternalGeneIdentifierOffsetBase]

    class Config:
        getter_dict: ExternalIdentifiersGetter


class TargetGeneModify(TargetGeneBase):
    pass


class TargetGeneCreate(TargetGeneModify):
    """View model for creating a new target gene."""

    target_sequence: Optional[TargetSequenceCreate]
    target_accession: Optional[TargetAccessionCreate]
    external_identifiers: Sequence[external_gene_identifier_offset.ExternalGeneIdentifierOffsetCreate]

    @root_validator()
    def sequence_or_accession_required_and_mutually_exclusive(cls, values: dict[str, Any]) -> dict[str, Any]:
        target_seq, target_acc = values.get("target_sequence"), values.get("target_accession")
        if target_seq is not None and target_acc is not None:
            raise ValueError("Expected either a `target_sequence` or a `target_accession`, not both.")
        if target_seq is None and target_acc is None:
            raise ValueError("Expected either a `target_sequence` or a `target_accession`, not neither.")

        return values


class TargetGeneUpdate(TargetGeneModify):
    """View model for updating a target gene."""

    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffsetCreate]


class SavedTargetGene(TargetGeneBase):
    """Base class for target gene view models representing saved records."""

    id: int
    record_type: str = None  # type: ignore
    target_sequence: Optional[SavedTargetSequence]
    target_accession: Optional[SavedTargetAccession]
    external_identifiers: Sequence[external_gene_identifier_offset.SavedExternalGeneIdentifierOffset]

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


class TargetGene(SavedTargetGene):
    """Target gene view model containing a complete set of properties visible to non-admin users."""

    target_sequence: Optional[TargetSequence]
    target_accession: Optional[TargetAccession]
    external_identifiers: Sequence[external_gene_identifier_offset.ExternalGeneIdentifierOffset]

    class Config:
        getter_dict = ExternalIdentifiersGetter

    @validator("target_accession", always=True)
    def check_seq_or_accession(cls, target_accession, values):
        if "target_sequence" not in values and not target_accession:
            raise ValueError("either a `target_sequence` or `target_accession` is required")
        return target_accession


class ShortTargetGene(SavedTargetGene):
    """Target gene view model containing a smaller set of properties to return in list contexts."""

    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffset]

    class Config:
        getter_dict = ExternalIdentifiersGetter


class AdminTargetGene(SavedTargetGene):
    """Target gene view model containing properties to return to admin clients."""

    creation_date: date
    modification_date: date
    target_sequence: Optional[AdminTargetSequence]
    target_accession: Optional[TargetAccession]
    external_identifiers: Sequence[external_gene_identifier_offset.ExternalGeneIdentifierOffset]

    class Config:
        getter_dict = ExternalIdentifiersGetter

    @validator("target_accession", always=True)
    def check_seq_or_accession(cls, target_accession, values):
        if "target_sequence" not in values and not target_accession:
            raise ValueError("either a `target_sequence` or `target_accession` is required")
        return target_accession



import os

import yaml
from pkg_resources import resource_stream


def load_stock_config(name="default"):
    """
    Loads a built-in stock logging configuration based on *name*.
    """
    with resource_stream(__package__, f"configurations/{name}.yaml") as file:
        return load_config(file)


def load_config(config):
    """
    Loads a given logging *config* written in YAML.

    *config* may be a string or open file object, both of which are accepted as
    the first argument to :py:func:`yaml.load`.
    """
    return yaml.load(config, Loader=LogConfigLoader)


class LogConfigLoader(yaml.SafeLoader):
    """
    A :py:class:`yaml.SafeLoader` subclass which implements some custom `!` tags.

    Local, custom tags supported:

    * ``!LOG_LEVEL``
    * ``!coalesce``

    >>> os.environ["LOG_LEVEL"] = "info"
    >>> yaml.load("level: !LOG_LEVEL", Loader = LogConfigLoader)
    {'level': 'INFO'}

    >>> os.environ["LOG_LEVEL"] = ""
    >>> yaml.load("level: !LOG_LEVEL", Loader = LogConfigLoader)
    {'level': None}
    >>> yaml.load('''
    ... level: !coalesce
    ...   - !LOG_LEVEL
    ...   - WARNING
    ... ''', Loader = LogConfigLoader)
    {'level': 'WARNING'}

    >>> del os.environ["LOG_LEVEL"]
    >>> yaml.load("level: !LOG_LEVEL", Loader = LogConfigLoader)
    {'level': None}
    >>> yaml.load('''
    ... level: !coalesce
    ...   - !LOG_LEVEL
    ...   - WARNING
    ... ''', Loader = LogConfigLoader)
    {'level': 'WARNING'}
    """

    pass


def log_level_constructor(loader, node):
    """
    Implements a custom YAML tag ``!LOG_LEVEL``.

    Produces the uppercased value of the ``LOG_LEVEL`` environment variable,
    if the variable has a value.  Otherwise, returns ``None``.
    """
    level = os.environ.get("LOG_LEVEL")
    return level.upper() if level else None


def cloud_watch_log_group_constructor(loader, node):
    """
    Implements a custom YAML tag ``!CLOUD_WATCH_LOG_GROUP``.

    Produces the value of the ``CLOUD_WATCH_LOG_GROUP`` environment variable,
    if the variable has a value.  Otherwise, returns ``None``.
    """
    group = os.environ.get("CLOUD_WATCH_LOG_GROUP")
    return group if group else None


def use_queues_constructor(loader, node):
    """
    Implements a custom YAML tag ``!CLOUD_WATCH_USE_QUEUES``.

    Produces the value of the ``CLOUD_WATCH_USE_QUEUES`` environment variable,
    if the variable has a value.  Otherwise, returns ``true`` to maintain defaults.
    """
    use_queues = os.environ.get("CLOUD_WATCH_USE_QUEUES")
    return use_queues.lower() == "true" if use_queues else False


def coalesce_constructor(loader, node):
    """
    Implements a custom YAML tag ``!coalesce``.

    When applied to a YAML sequence (e.g. list or array), the produced value is
    the first value which is not ``None``.  Akin to SQL's coalesce() function.
    """
    values = loader.construct_sequence(node)
    return first(lambda x: x is not None, values)


def first(predicate, iterable):
    """
    Return the first item in *iterable* for which *predicate* returns ``True``.

    If *iterable* is empty or contains no items passing *predicate*, returns
    ``None``.
    """
    return next(filter(predicate, iterable), None)


LogConfigLoader.add_constructor("!LOG_LEVEL", log_level_constructor)
LogConfigLoader.add_constructor("!CLOUD_WATCH_LOG_GROUP", cloud_watch_log_group_constructor)
LogConfigLoader.add_constructor("!CLOUD_WATCH_USE_QUEUES", use_queues_constructor)
LogConfigLoader.add_constructor("!coalesce", coalesce_constructor)



from typing import Optional

from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.utilities import is_null


# TODO: value will not be Optional when we confirm the final controlled keyword list.
def validate_description(value: str, key: str, description: Optional[str]):
    if value.lower() == "other" and (description is None or description.strip() == ""):
        raise ValidationError(
            "Other option does not allow empty description.", custom_loc=["body", "keywordDescriptions", key]
        )


def validate_duplicates(keywords: list):
    keys = []
    values = []
    for k in keywords:
        keys.append(k.keyword.key.lower())  # k: ExperimentControlledKeywordCreate object
        if k.keyword.value.lower() != "other":
            values.append(k.keyword.value.lower())

    keys_set = set(keys)
    values_set = set(values)

    if len(keys) != len(keys_set):
        raise ValidationError("Duplicate keys found in keywords.")
    if len(values) != len(values_set):
        raise ValidationError("Duplicate values found in keywords.")


def validate_keyword(keyword: str):
    """
    Validates a keyword.

    Parameters
    __________
    keyword: str

    Raises
    ______
    ValidationError
        If keyword is invalid or null.
    """
    if is_null(keyword) or not isinstance(keyword, str):
        raise ValidationError("{} are not valid keyword. Keyword must be a non null list of strings.".format(keyword))


def validate_keyword_keys(keywords: list):
    keyword_dict = {k.keyword.key.lower(): k.keyword.value.lower() for k in keywords}
    variant_library_method = keyword_dict.get("variant library creation method", "")

    if variant_library_method == "endogenous locus library method":
        if not (
            "endogenous locus library method system" in keyword_dict
            and "endogenous locus library method mechanism" in keyword_dict
        ):
            raise ValidationError(
                "If 'Variant Library Creation Method' is 'Endogenous locus library method', "
                "both 'Endogenous Locus Library Method System' and 'Endogenous Locus Library Method Mechanism' "
                "must be present."
            )

    elif variant_library_method == "in vitro construct library method":
        if not (
            "in vitro construct library method system" in keyword_dict
            and "in vitro construct library method mechanism" in keyword_dict
        ):
            raise ValidationError(
                "If 'Variant Library Creation Method' is 'In vitro construct library method', "
                "both 'In Vitro Construct Library Method System' and 'In Vitro Construct Library Method Mechanism' "
                "must be present."
            )

    elif variant_library_method == "other":
        if any(
            k in keyword_dict
            for k in [
                "endogenous locus library method system",
                "endogenous locus library method mechanism",
                "in vitro construct library method system",
                "in vitro construct library method mechanism",
            ]
        ):
            raise ValidationError(
                "If 'Variant Library Creation Method' is 'Other', none of "
                "'Endogenous Locus Library Method System', 'Endogenous Locus Library Method Mechanism', "
                "'In Vitro Construct Library Method System', or 'In Vitro Construct Library Method Mechanism' "
                "should be present."
            )


def validate_keyword_list(keywords: list):
    validate_duplicates(keywords)
    validate_keyword_keys(keywords)



import os
from datetime import date
from typing import Optional, TypedDict

import requests
from cdot.hgvs.dataproviders import ChainedSeqFetcher, FastaSeqFetcher, RESTDataProvider

GENOMIC_FASTA_FILES = [
    "/data/GCF_000001405.39_GRCh38.p13_genomic.fna.gz",
    "/data/GCF_000001405.25_GRCh37.p13_genomic.fna.gz",
]

DCD_MAP_URL = os.environ.get("DCD_MAPPING_URL", "http://dcd-mapping:8000")


def seqfetcher() -> ChainedSeqFetcher:
    return ChainedSeqFetcher(*[FastaSeqFetcher(file) for file in GENOMIC_FASTA_FILES])


def cdot_rest() -> RESTDataProvider:
    return RESTDataProvider(seqfetcher=seqfetcher())


class VRSMap:
    url: str

    class ScoreSetMappingResults(TypedDict):
        metadata: Optional[dict[str, str]]
        dcd_mapping_version: str
        mapped_date_utc: date
        computed_genomic_reference_sequence: Optional[dict[str, str]]
        mapped_genomic_reference_sequence: Optional[dict[str, str]]
        computed_protein_reference_sequence: Optional[dict[str, str]]
        mapped_protein_reference_sequence: Optional[dict[str, str]]
        mapped_scores: Optional[list[dict]]
        error_message: Optional[str]

    def __init__(self, url: str) -> None:
        self.url = url

    def map_score_set(self, score_set_urn: str) -> ScoreSetMappingResults:
        uri = f"{self.url}/api/v1/map/{score_set_urn}"
        response = requests.post(uri)
        response.raise_for_status()
        return response.json()


def vrs_mapper(url: Optional[str] = None) -> VRSMap:
    return VRSMap(DCD_MAP_URL) if not url else VRSMap(url)



import os
from concurrent import futures
from typing import Callable

from arq.connections import RedisSettings
from arq.cron import CronJob

from mavedb.data_providers.services import cdot_rest
from mavedb.db.session import SessionLocal
from mavedb.lib.logging.canonical import log_job
from mavedb.worker.jobs import create_variants_for_score_set, map_variants_for_score_set, variant_mapper_manager

# ARQ requires at least one task on startup.
BACKGROUND_FUNCTIONS: list[Callable] = [
    create_variants_for_score_set,
    variant_mapper_manager,
    map_variants_for_score_set,
]
BACKGROUND_CRONJOBS: list[CronJob] = []

REDIS_IP = os.getenv("REDIS_IP") or "localhost"
REDIS_PORT = int(os.getenv("REDIS_PORT") or 6379)
REDIS_SSL = (os.getenv("REDIS_SSL") or "false").lower() == "true"


RedisWorkerSettings = RedisSettings(host=REDIS_IP, port=REDIS_PORT, ssl=REDIS_SSL)


async def startup(ctx):
    ctx["pool"] = futures.ProcessPoolExecutor()


async def shutdown(ctx):
    pass


async def on_job_start(ctx):
    db = SessionLocal()
    db.current_user_id = None
    ctx["db"] = db
    ctx["hdp"] = cdot_rest()
    ctx["state"] = {}


async def on_job_end(ctx):
    db = ctx["db"]
    db.close()


class ArqWorkerSettings:
    """
    Settings for the ARQ worker.
    """

    on_startup = startup
    on_shutdown = shutdown
    on_job_start = on_job_start
    on_job_end = on_job_end
    after_job_end = log_job
    redis_settings = RedisWorkerSettings
    functions: list = BACKGROUND_FUNCTIONS
    cron_jobs: list = BACKGROUND_CRONJOBS

    job_timeout = 5 * 60 * 60  # Keep jobs alive for a long while...



from typing import Optional

from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel


class ContributorBase(BaseModel):
    """Base class for contributor view models."""

    orcid_id: str


class ContributorCreate(ContributorBase):
    """View model for creating a new contributor or looking one up."""

    pass


class SavedContributor(ContributorBase):
    """Base class for contributor view models representing saved records."""

    record_type: str = None  # type: ignore
    given_name: Optional[str]
    family_name: Optional[str]

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True


class Contributor(SavedContributor):
    """Contributor view model."""

    pass



from sqlalchemy.orm import configure_mappers

from mavedb.models import *  # noqa: F403
from mavedb.worker.settings import ArqWorkerSettings

# Scan all our model classes and create backref attributes. Otherwise, these attributes only get added to classes once
# an instance of the related class has been created. Since the worker is a distinct service, we should make sure this
# happens whenever a worker is spawned.
configure_mappers()

# Expose the worker settings to the arq CLI Worker initializer
global WorkerSettings
WorkerSettings = ArqWorkerSettings



import logging
from enum import Enum
from typing import Optional

from mavedb.db.base import Base
from mavedb.lib.authentication import UserData
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.models.enums.user_role import UserRole
from mavedb.models.experiment import Experiment
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.score_set import ScoreSet
from mavedb.models.user import User

logger = logging.getLogger(__name__)


class Action(Enum):
    READ = "read"
    UPDATE = "update"
    DELETE = "delete"
    ADD_EXPERIMENT = "add_experiment"
    ADD_SCORE_SET = "add_score_set"
    SET_SCORES = "set_scores"
    ADD_ROLE = "add_role"
    PUBLISH = "publish"


class PermissionResponse:
    def __init__(self, permitted: bool, http_code: int = 403, message: Optional[str] = None):
        self.permitted = permitted
        self.http_code = http_code if not permitted else None
        self.message = message if not permitted else None

        save_to_logging_context({"permission_message": self.message, "access_permitted": self.permitted})
        if self.permitted:
            logger.debug(msg="Access to the requested resource is permitted.", extra=logging_context())
        else:
            logger.debug(msg="Access to the requested resource is not permitted.", extra=logging_context())


class PermissionException(Exception):
    def __init__(self, http_code: int, message: str):
        self.http_code = http_code
        self.message = message


def roles_permitted(user_roles: list[UserRole], permitted_roles: list[UserRole]) -> bool:
    save_to_logging_context({"permitted_roles": [role.name for role in permitted_roles]})

    if not user_roles:
        logger.debug(msg="User has no associated roles.", extra=logging_context())
        return False

    return any(role in permitted_roles for role in user_roles)


def has_permission(user_data: Optional[UserData], item: Base, action: Action) -> PermissionResponse:
    private = False
    user_is_owner = False
    user_is_self = False
    user_may_edit = False
    active_roles = user_data.active_roles if user_data else []

    if isinstance(item, ExperimentSet) or isinstance(item, Experiment) or isinstance(item, ScoreSet):
        assert item.private is not None
        private = item.private
        published = item.published_date is not None
        user_is_owner = item.created_by_id == user_data.user.id if user_data is not None else False
        user_may_edit = user_is_owner or (
            user_data is not None and user_data.user.username in [c.orcid_id for c in item.contributors]
        )

        save_to_logging_context({"resource_is_published": published})

    if isinstance(item, User):
        user_is_self = item.id == user_data.user.id if user_data is not None else False
        user_may_edit = user_is_self

    save_to_logging_context(
        {
            "resource_is_private": private,
            "user_is_owner_of_resource": user_is_owner,
            "user_is_may_edit_resource": user_may_edit,
            "user_is_self": user_is_self,
        }
    )

    if isinstance(item, ExperimentSet):
        if action == Action.READ:
            if user_may_edit or not private:
                return PermissionResponse(True)
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin, UserRole.mapper]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"experiment set with URN '{item.urn}' not found")
            elif user_data is None or user_data.user is None:
                return PermissionResponse(False, 401, f"insufficient permissions for URN '{item.urn}'")
            else:
                return PermissionResponse(False, 403, f"insufficient permissions for URN '{item.urn}'")
        elif action == Action.UPDATE:
            if user_may_edit:
                return PermissionResponse(True)
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"experiment set with URN '{item.urn}' not found")
            elif user_data is None or user_data.user is None:
                return PermissionResponse(False, 401, f"insufficient permissions for URN '{item.urn}'")
            else:
                return PermissionResponse(False, 403, f"insufficient permissions for URN '{item.urn}'")
        elif action == Action.DELETE:
            # Owner may only delete an experiment set if it has not already been published.
            if user_may_edit:
                return PermissionResponse(not published, 403, f"insufficient permissions for URN '{item.urn}'")
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"experiment set with URN '{item.urn}' not found")
            else:
                return PermissionResponse(False)
        elif action == Action.ADD_EXPERIMENT:
            # Only permitted users can add an experiment to an existing experiment set.
            return PermissionResponse(
                user_may_edit or roles_permitted(active_roles, [UserRole.admin]),
                404 if private else 403,
                (
                    f"experiment set with URN '{item.urn}' not found"
                    if private
                    else f"insufficient permissions for URN '{item.urn}'"
                ),
            )
        else:
            raise NotImplementedError(f"has_permission(User, ExperimentSet, {action}, Role)")

    elif isinstance(item, Experiment):
        if action == Action.READ:
            if user_may_edit or not private:
                return PermissionResponse(True)
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin, UserRole.mapper]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"experiment with URN '{item.urn}' not found")
            elif user_data is None or user_data.user is None:
                return PermissionResponse(False, 401, f"insufficient permissions for URN '{item.urn}'")
            else:
                return PermissionResponse(False, 403, f"insufficient permissions for URN '{item.urn}'")
        elif action == Action.UPDATE:
            if user_may_edit:
                return PermissionResponse(True)
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"experiment with URN '{item.urn}' not found")
            elif user_data is None or user_data.user is None:
                return PermissionResponse(False, 401, f"insufficient permissions for URN '{item.urn}'")
            else:
                return PermissionResponse(False, 403, f"insufficient permissions for URN '{item.urn}'")
        elif action == Action.DELETE:
            # Owner may only delete an experiment if it has not already been published.
            if user_may_edit:
                return PermissionResponse(not published, 403, f"insufficient permissions for URN '{item.urn}'")
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"experiment set with URN '{item.urn}' not found")
            else:
                return PermissionResponse(False)
        elif action == Action.ADD_SCORE_SET:
            # Only permitted users can add a score set to a private experiment.
            if user_may_edit or roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            elif private:
                return PermissionResponse(False, 404, f"experiment with URN '{item.urn}' not found")
            # Any signed in user has permissions to add a score set to a public experiment
            elif user_data is not None:
                return PermissionResponse(True)
            else:
                return PermissionResponse(False, 403, f"insufficient permissions for URN '{item.urn}'")
        else:
            raise NotImplementedError(f"has_permission(User, Experiment, {action}, Role)")

    elif isinstance(item, ScoreSet):
        if action == Action.READ:
            if user_may_edit or not private:
                return PermissionResponse(True)
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin, UserRole.mapper]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"score set with URN '{item.urn}' not found")
            elif user_data is None or user_data.user is None:
                return PermissionResponse(False, 401, f"insufficient permissions for URN '{item.urn}'")
            else:
                return PermissionResponse(False, 403, f"insufficient permissions for URN '{item.urn}'")
        elif action == Action.UPDATE:
            if user_may_edit:
                return PermissionResponse(True)
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"score set with URN '{item.urn}' not found")
            elif user_data is None or user_data.user is None:
                return PermissionResponse(False, 401, f"insufficient permissions for URN '{item.urn}'")
            else:
                return PermissionResponse(False, 403, f"insufficient permissions for URN '{item.urn}'")
        elif action == Action.DELETE:
            # Owner may only delete a score set if it has not already been published.
            if user_may_edit:
                return PermissionResponse(not published, 403, f"insufficient permissions for URN '{item.urn}'")
            # Roles which may perform this operation.
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"experiment set with URN '{item.urn}' not found")
            else:
                return PermissionResponse(False)
        # Only the owner may publish a private score set.
        elif action == Action.PUBLISH:
            if user_may_edit:
                return PermissionResponse(True)
            elif roles_permitted(active_roles, []):
                return PermissionResponse(True)
            elif private:
                # Do not acknowledge the existence of a private entity.
                return PermissionResponse(False, 404, f"score set with URN '{item.urn}' not found")
            else:
                return PermissionResponse(False)
        elif action == Action.SET_SCORES:
            return PermissionResponse(
                (user_may_edit or roles_permitted(active_roles, [UserRole.admin])),
                404 if private else 403,
                (
                    f"score set with URN '{item.urn}' not found"
                    if private
                    else f"insufficient permissions for URN '{item.urn}'"
                ),
            )
        else:
            raise NotImplementedError(f"has_permission(User, ScoreSet, {action}, Role)")

    elif isinstance(item, User):
        if action == Action.READ:
            if user_is_self:
                return PermissionResponse(True)
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            else:
                return PermissionResponse(False, 403, "Insufficient permissions for user update.")
        elif action == Action.UPDATE:
            if user_is_self:
                return PermissionResponse(True)
            elif roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            else:
                return PermissionResponse(False, 403, "Insufficient permissions for user update.")
        elif action == Action.ADD_ROLE:
            if roles_permitted(active_roles, [UserRole.admin]):
                return PermissionResponse(True)
            else:
                return PermissionResponse(False, 403, "Insufficient permissions to add user role.")
        elif action == Action.DELETE:
            raise NotImplementedError(f"has_permission(User, ScoreSet, {action}, Role)")
        else:
            raise NotImplementedError(f"has_permission(User, ScoreSet, {action}, Role)")

    else:
        raise NotImplementedError(f"has_permission(User, {item.__class__}, {action}, Role)")


def assert_permission(user_data: Optional[UserData], item: Base, action: Action) -> PermissionResponse:
    save_to_logging_context({"permission_boundary": action.name})
    permission = has_permission(user_data, item, action)

    if not permission.permitted:
        assert permission.http_code and permission.message
        raise PermissionException(http_code=permission.http_code, message=permission.message)

    return permission



import logging
from operator import attrgetter
from typing import Any, Optional

import pydantic
import requests
from fastapi import APIRouter, Depends, HTTPException
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import UserData, get_current_user
from mavedb.lib.authorization import require_current_user, require_current_user_with_email
from mavedb.lib.contributors import find_or_create_contributor
from mavedb.lib.exceptions import NonexistentOrcidUserError
from mavedb.lib.experiments import search_experiments as _search_experiments
from mavedb.lib.identifiers import (
    find_or_create_doi_identifier,
    find_or_create_publication_identifier,
    find_or_create_raw_read_identifier,
)
from mavedb.lib.keywords import search_keyword
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.permissions import Action, assert_permission, has_permission
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.keywords import validate_keyword_list
from mavedb.models.contributor import Contributor
from mavedb.models.experiment import Experiment
from mavedb.models.experiment_controlled_keyword import ExperimentControlledKeywordAssociation
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.score_set import ScoreSet
from mavedb.view_models import experiment, score_set
from mavedb.view_models.search import ExperimentsSearch

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/api/v1",
    tags=["experiments"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)


# TODO: Rewrite this function.
@router.get(
    "/experiments/",
    status_code=200,
    response_model=list[experiment.Experiment],
    response_model_exclude_none=True,
)
def list_experiments(
    *,
    editable: Optional[bool] = None,
    q: Optional[str] = None,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> list[Experiment]:
    """
    List experiments.
    """
    query = db.query(Experiment)
    if q is not None:
        save_to_logging_context({"query_string": q})

        if user_data is None or user_data.user is None:
            logger.debug(msg="User is anonymous; Cannot list their experiments.", extra=logging_context())
            return []

        if len(q) > 0:
            logger.debug(msg="Listing experiments for the current user.", extra=logging_context())
            query = query.filter(
                Experiment.created_by_id == user_data.user.id
            )  # .filter(Experiment.published_date is None)
        # else:
        #     query = query.filter(Experiment.created_by_id == user.id).filter(Experiment.published_date is None)
    else:
        logger.debug(msg="No query string was provided; Listing all experiments.", extra=logging_context())

    items = query.order_by(Experiment.urn).all()
    return items


@router.post(
    "/experiments/search",
    status_code=200,
    response_model=list[experiment.ShortExperiment],
)
def search_experiments(search: ExperimentsSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search experiments.
    """
    return _search_experiments(db, None, search)


@router.post(
    "/me/experiments/search",
    status_code=200,
    response_model=list[experiment.ShortExperiment],
)
def search_my_experiments(
    search: ExperimentsSearch,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Search experiments created by the current user..
    """
    return _search_experiments(db, user_data.user, search)


@router.get(
    "/experiments/{urn}",
    status_code=200,
    response_model=experiment.Experiment,
    responses={404: {}},
    response_model_exclude_none=True,
)
def fetch_experiment(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> Experiment:
    """
    Fetch a single experiment by URN.
    """
    # item = db.query(Experiment).filter(Experiment.urn == urn).filter(Experiment.private.is_(False)).first()
    item = db.query(Experiment).filter(Experiment.urn == urn).first()
    save_to_logging_context({"requested_resource": urn})

    if not item:
        logger.debug(msg="The requested experiment does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"Experiment with URN {urn} not found")

    assert_permission(user_data, item, Action.READ)
    return item


@router.get(
    "/experiments/{urn}/score-sets",
    status_code=200,
    response_model=list[score_set.ScoreSet],
    responses={404: {}},
    response_model_exclude_none=True,
)
def get_experiment_score_sets(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: Optional[UserData] = Depends(get_current_user),
) -> Any:
    """
    Get all score sets belonging to an experiment.
    """
    save_to_logging_context({"requested_resource": urn, "resource_property": "score-sets"})

    experiment = db.query(Experiment).filter(Experiment.urn == urn).first()
    if not experiment:
        logger.debug(msg="The requested experiment does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"experiment with URN '{urn}' not found")

    assert_permission(user_data, experiment, Action.READ)

    score_set_result = (
        db.query(ScoreSet)
        .filter(ScoreSet.experiment_id == experiment.id)
        .filter(~ScoreSet.superseding_score_set.has())
        .all()
    )
    score_set_result[:] = [
        score_set for score_set in score_set_result if has_permission(user_data, score_set, Action.READ).permitted
    ]

    if not score_set_result:
        save_to_logging_context({"associated_resources": []})
        logger.info(msg="No score sets are associated with the requested experiment.", extra=logging_context())

        raise HTTPException(status_code=404, detail="no associated score sets")
    else:
        score_set_result.sort(key=attrgetter("urn"))
        save_to_logging_context({"associated_resources": [item.urn for item in score_set_result]})

    return score_set_result


@router.post(
    "/experiments/",
    response_model=experiment.Experiment,
    responses={422: {}},
    response_model_exclude_none=True,
)
async def create_experiment(
    *,
    item_create: experiment.ExperimentCreate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Create an experiment.
    """
    logger.debug(msg="Began creation of new experiment.", extra=logging_context())

    experiment_set = None
    if item_create.experiment_set_urn is not None:
        experiment_set = (
            db.query(ExperimentSet).filter(ExperimentSet.urn == item_create.experiment_set_urn).one_or_none()
        )
        if not experiment_set:
            logger.info(
                msg="Could not create experiment; The requested experiment set does not exist.", extra=logging_context()
            )
            raise HTTPException(
                status_code=404,
                detail=f"experiment set with URN '{item_create.experiment_set_urn}' not found.",
            )

        save_to_logging_context({"experiment_set": experiment_set.urn})
        assert_permission(user_data, experiment_set, Action.ADD_EXPERIMENT)

        logger.debug(msg="Creating experiment within existing experiment set.", extra=logging_context())

    contributors: list[Contributor] = []
    try:
        contributors = [
            await find_or_create_contributor(db, contributor.orcid_id) for contributor in item_create.contributors or []
        ]
    except NonexistentOrcidUserError as e:
        logger.error(msg="Could not find ORCID user with the provided user ID.", extra=logging_context())
        raise pydantic.ValidationError(
            [pydantic.error_wrappers.ErrorWrapper(ValidationError(str(e)), loc="contributors")],
            model=experiment.ExperimentCreate,
        )

    try:
        doi_identifiers = [
            await find_or_create_doi_identifier(db, identifier.identifier)
            for identifier in item_create.doi_identifiers or []
        ]
        raw_read_identifiers = [
            await find_or_create_raw_read_identifier(db, identifier.identifier)
            for identifier in item_create.raw_read_identifiers or []
        ]
        primary_publication_identifiers = [
            await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
            for identifier in item_create.primary_publication_identifiers or []
        ]
        publication_identifiers = [
            await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
            for identifier in item_create.secondary_publication_identifiers or []
        ] + primary_publication_identifiers

    except requests.exceptions.ConnectTimeout:
        logger.error(msg="Gateway timed out while creating experiment identifiers.", extra=logging_context())
        raise HTTPException(status_code=504, detail="Gateway Timeout")

    except requests.exceptions.HTTPError:
        logger.error(msg="Encountered bad gateway while creating experiment identifiers.", extra=logging_context())
        raise HTTPException(status_code=502, detail="Bad Gateway")

    # create a temporary `primary` attribute on each of our publications that indicates
    # to our association proxy whether it is a primary publication or not
    primary_identifiers = [pub.identifier for pub in primary_publication_identifiers]
    for publication in publication_identifiers:
        setattr(publication, "primary", publication.identifier in primary_identifiers)

    # TODO: Controlled keywords currently is allowed none value.
    #  Will be changed in the future when we get the final list.
    keywords: list[ExperimentControlledKeywordAssociation] = []
    if item_create.keywords:
        all_values_none = all(k.keyword.value is None for k in item_create.keywords)
        if all_values_none is False:
            # Users may choose part of keywords from dropdown menu. Remove not chosen keywords from the list.
            filtered_keywords = list(filter(lambda k: k.keyword.value is not None, item_create.keywords))
            try:
                validate_keyword_list(filtered_keywords)
            except ValidationError as e:
                raise HTTPException(status_code=422, detail=str(e))
            for upload_keyword in filtered_keywords:
                try:
                    description = upload_keyword.description
                    controlled_keyword = search_keyword(db, upload_keyword.keyword.key, upload_keyword.keyword.value)
                    experiment_controlled_keyword = ExperimentControlledKeywordAssociation(
                        controlled_keyword=controlled_keyword,
                        description=description,
                    )
                    keywords.append(experiment_controlled_keyword)
                except ValueError as e:
                    raise HTTPException(status_code=422, detail=str(e))

    item = Experiment(
        **jsonable_encoder(
            item_create,
            by_alias=False,
            exclude={
                "contributors",
                "doi_identifiers",
                "experiment_set_urn",
                "keywords",
                "primary_publication_identifiers",
                "secondary_publication_identifiers",
                "raw_read_identifiers",
            },
        ),
        experiment_set=experiment_set,
        contributors=contributors,
        doi_identifiers=doi_identifiers,
        publication_identifiers=publication_identifiers,  # an internal association proxy representation
        raw_read_identifiers=raw_read_identifiers,
        created_by=user_data.user,
        modified_by=user_data.user,
        keyword_objs=keywords,
    )  # type: ignore

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"created_resource": item.urn})
    return item


@router.put(
    "/experiments/{urn}",
    response_model=experiment.Experiment,
    responses={422: {}},
    response_model_exclude_none=True,
)
async def update_experiment(
    *,
    item_update: experiment.ExperimentUpdate,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user_with_email),
) -> Any:
    """
    Update an experiment.
    """
    save_to_logging_context({"requested_resource": urn})
    logger.debug(msg="Began experiment update.", extra=logging_context())

    # item = db.query(Experiment).filter(Experiment.urn == urn).filter(Experiment.private.is_(False)).one_or_none()
    item = db.query(Experiment).filter(Experiment.urn == urn).one_or_none()
    if item is None:
        logger.info(
            msg="Failed to update experiment; The requested experiment does not exist.", extra=logging_context()
        )
        raise HTTPException(status_code=404, detail=f"experiment with URN {urn} not found")

    assert_permission(user_data, item, Action.UPDATE)

    pairs = {
        k: v
        for k, v in vars(item_update).items()
        if k
        not in [
            "contributors",
            "doi_identifiers",
            "keywords",
            "secondary_publication_identifiers",
            "primary_publication_identifiers",
            "raw_read_identifiers",
        ]
    }
    for var, value in pairs.items():  # vars(item_update).items():
        setattr(item, var, value) if value else None

    try:
        item.contributors = [
            await find_or_create_contributor(db, contributor.orcid_id) for contributor in item_update.contributors or []
        ]
    except NonexistentOrcidUserError as e:
        logger.error(msg="Could not find ORCID user with the provided user ID.", extra=logging_context())
        raise pydantic.ValidationError(
            [pydantic.error_wrappers.ErrorWrapper(ValidationError(str(e)), loc="contributors")],
            model=experiment.ExperimentUpdate,
        )

    doi_identifiers = [
        await find_or_create_doi_identifier(db, identifier.identifier)
        for identifier in item_update.doi_identifiers or []
    ]
    raw_read_identifiers = [
        await find_or_create_raw_read_identifier(db, identifier.identifier)
        for identifier in item_update.raw_read_identifiers or []
    ]

    primary_publication_identifiers = [
        await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
        for identifier in item_update.primary_publication_identifiers or []
    ]
    publication_identifiers = [
        await find_or_create_publication_identifier(db, identifier.identifier, identifier.db_name)
        for identifier in item_update.secondary_publication_identifiers or []
    ] + primary_publication_identifiers

    # create a temporary `primary` attribute on each of our publications that indicates
    # to our association proxy whether it is a primary publication or not
    primary_identifiers = [pub.identifier for pub in primary_publication_identifiers]
    for publication in publication_identifiers:
        setattr(publication, "primary", publication.identifier in primary_identifiers)

    item.doi_identifiers = doi_identifiers
    item.publication_identifiers = publication_identifiers
    item.raw_read_identifiers = raw_read_identifiers

    if item_update.keywords:
        all_values_none = all(k.keyword.value is None for k in item_update.keywords)
        if all_values_none is False:
            # Users may choose part of keywords from dropdown menu. Remove not chosen keywords from the list.
            filtered_keywords = list(filter(lambda k: k.keyword.value is not None, item_update.keywords))
            try:
                validate_keyword_list(filtered_keywords)
            except ValidationError as e:
                raise HTTPException(status_code=422, detail=str(e))
            try:
                await item.set_keywords(db, filtered_keywords)
            except Exception as e:
                raise HTTPException(status_code=500, detail=f"Invalid keywords: {str(e)}")

    item.modified_by = user_data.user

    db.add(item)
    db.commit()
    db.refresh(item)

    save_to_logging_context({"updated_resource": item.urn})
    return item


@router.delete("/experiments/{urn}", response_model=None, responses={422: {}})
async def delete_experiment(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> None:
    """
    Delete a experiment .

    Raises

    Returns
    _______
    Does not return anything
    string : HTTP code 200 successful but returning content
    or
    communitcate to client whether the operation succeeded
    204 if successful but not returning content - likely going with this
    """
    save_to_logging_context({"requested_resource": urn})

    item = db.query(Experiment).filter(Experiment.urn == urn).one_or_none()
    if not item:
        logger.info(
            msg="Could not delete the requested experiment; The requested experiment does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"experiment with URN '{urn}' not found.")

    assert_permission(user_data, item, Action.DELETE)

    db.delete(item)
    db.commit()



import json
import os
import sys
import traceback

from slack_sdk.webhook import WebhookClient


def find_traceback_locations():
    _, _, tb = sys.exc_info()
    return [
        (fs.filename, fs.lineno, fs.name)
        for fs in traceback.extract_tb(tb)
        # attempt to show only *our* code, not the many layers of library code
        if "/mavedb/" in fs.filename and "/.direnv/" not in fs.filename
    ]


def send_slack_message(err, request=None):
    text = {"type": err.__class__.__name__, "exception": str(err), "location": find_traceback_locations()}

    if request:
        text["client"] = str(request.client.host)
        text["request"] = f"{request.method} {request.url}"

    text = json.dumps(text)
    slack_webhook_url = os.getenv("SLACK_WEBHOOK_URL")
    if slack_webhook_url is not None and len(slack_webhook_url) > 0:
        client = WebhookClient(url=slack_webhook_url)
        client.send(
            text=text,
            blocks=[
                {
                    "type": "section",
                    "text": {"type": "plain_text", "text": text},
                }
            ],
        )
    else:
        print(f"EXCEPTION_HANDLER: {text}")



from typing import Any, Dict, Union

from sqlalchemy.orm import as_declarative, declared_attr

class_registry: Dict = {}


@as_declarative(class_registry=class_registry)
class Base:
    id: Any
    __name__: str

    # Generate __tablename__ automatically
    # Declared in this odd way to provide correct type hint for mypy
    __tablename__: Union[declared_attr[Any], str] = declared_attr(lambda cls: cls.__name__.lower())



import logging
from typing import Optional

from mavedb.lib.identifiers import PublicationAuthors
from mavedb.lib.validation.publication import validate_db_name, validate_publication
from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel, validator

logger = logging.getLogger(__name__)


class PublicationIdentifierBase(BaseModel):
    identifier: str
    db_name: Optional[str]


class PublicationIdentifierCreate(PublicationIdentifierBase):
    @validator("identifier")
    def validate_publication(cls, v):
        validate_publication(identifier=v)
        return v

    @validator("db_name")
    def validate_publication_db(cls, v):
        validate_db_name(db_name=v)
        return v


# Properties of external publication identifiers
class ExternalPublicationIdentifier(PublicationIdentifierBase):
    record_type: str = None  # type: ignore
    title: str
    authors: list[PublicationAuthors]

    abstract: Optional[str]
    doi: Optional[str]
    publication_year: Optional[int]
    publication_journal: Optional[str]
    url: Optional[str]
    reference_html: Optional[str]

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True


# Properties shared by models stored in DB
class SavedPublicationIdentifier(ExternalPublicationIdentifier):
    id: int


class ValidatedPublicationIdentifier(PublicationIdentifierBase):
    valid_for: dict[str, bool]


# Properties to return to non-admin clients
class PublicationIdentifier(SavedPublicationIdentifier):
    pass



from datetime import date

from pydantic.types import Optional

from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel


class TaxonomyBase(BaseModel):
    tax_id: int
    organism_name: Optional[str]
    common_name: Optional[str]
    rank: Optional[str]
    has_described_species_name: Optional[bool]
    article_reference: Optional[str]
    genome_id: Optional[int]


class TaxonomyCreate(TaxonomyBase):
    pass


class TaxonomyUpdate(TaxonomyBase):
    pass


# Properties shared by models stored in DB
class SavedTaxonomy(TaxonomyBase):
    id: int
    record_type: str = None  # type: ignore
    url: str

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class Taxonomy(SavedTaxonomy):
    pass


# Properties to return to admin clients
class AdminTaxonomy(SavedTaxonomy):
    creation_date: date
    modification_date: date



from datetime import datetime
from typing import TYPE_CHECKING, Optional

from sqlalchemy import Column, Date, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, relationship

from mavedb.db.base import Base
from mavedb.models.role import Role
from mavedb.models.user import User

if TYPE_CHECKING:
    from mavedb.lib.authorization import UserRole


class AccessKey(Base):
    __tablename__ = "access_keys"

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=False)
    user: Mapped[User] = relationship(back_populates="access_keys")
    role_id = Column(Integer, ForeignKey("roles.id"), nullable=True)
    role_obj: Mapped[Role] = relationship(Role)
    key_id = Column(String, unique=True, index=True, nullable=False)
    public_key = Column(String, nullable=False)
    name = Column(String, nullable=True)
    expiration_date = Column(Date, nullable=True)
    creation_time = Column(DateTime, nullable=True, default=datetime.now)

    @property
    def role(self) -> Optional["UserRole"]:
        role_obj = self.role_obj or None
        return None if role_obj is None else role_obj.name

    async def set_role(self, db, role: "UserRole"):
        self.role_obj = await self._find_or_create_role(db, role)

    @staticmethod
    async def _find_or_create_role(db, role):
        role_obj = db.query(Role).filter(Role.name == role).one_or_none()
        if not role_obj:
            role_obj = Role(name=role)

        return role_obj



from enum import Enum
from typing import Union

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import Table, func, select
from sqlalchemy.orm import Session

from mavedb.deps import get_db
from mavedb.models.controlled_keyword import ControlledKeyword
from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.models.ensembl_identifier import EnsemblIdentifier
from mavedb.models.ensembl_offset import EnsemblOffset
from mavedb.models.experiment import (
    Experiment,
    experiments_doi_identifiers_association_table,
    experiments_raw_read_identifiers_association_table,
)
from mavedb.models.experiment_controlled_keyword import ExperimentControlledKeywordAssociation
from mavedb.models.experiment_publication_identifier import ExperimentPublicationIdentifierAssociation
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.models.raw_read_identifier import RawReadIdentifier
from mavedb.models.refseq_identifier import RefseqIdentifier
from mavedb.models.refseq_offset import RefseqOffset
from mavedb.models.score_set import (
    ScoreSet,
    score_sets_doi_identifiers_association_table,
    score_sets_raw_read_identifiers_association_table,
)
from mavedb.models.score_set_publication_identifier import ScoreSetPublicationIdentifierAssociation
from mavedb.models.target_accession import TargetAccession
from mavedb.models.target_gene import TargetGene
from mavedb.models.target_sequence import TargetSequence
from mavedb.models.taxonomy import Taxonomy
from mavedb.models.uniprot_identifier import UniprotIdentifier
from mavedb.models.uniprot_offset import UniprotOffset
from mavedb.models.user import User

router = APIRouter(
    prefix="/api/v1/statistics",
    tags=["statistics"],
    responses={404: {"description": "Not found"}},
)

## Enum classes hold valid endpoints for different statistics routes.


class TargetGeneFields(str, Enum):
    category = "category"
    organism = "organism"

    ensemblIdentifier = "ensembl-identifier"
    refseqIdentifier = "refseq-identifier"
    uniprotIdentifier = "uniprot-identifier"


class TargetAccessionFields(str, Enum):
    accession = "accession"
    assembly = "assembly"
    gene = "gene"


class TargetSequenceFields(str, Enum):
    sequence = "sequence"
    sequenceType = "sequence-type"


class RecordNames(str, Enum):
    experiment = "experiment"
    scoreSet = "score-set"


class RecordFields(str, Enum):
    publicationIdentifiers = "publication-identifiers"
    keywords = "keywords"
    doiIdentifiers = "doi-identifiers"
    rawReadIdentifiers = "raw-read-identifiers"
    createdBy = "created-by"


def _target_from_field_and_model(
    db: Session,
    model: Union[type[TargetAccession], type[TargetSequence]],
    field: Union[TargetAccessionFields, TargetSequenceFields],
):
    """
    Given either the target accession or target sequence model, generate counts that can be used to create
    a statistic for those fields.
    """
    # Protection from this case occurs via FastApi/pydantic Enum validation on endpoints that reference this function.
    # If we are careful with our enumeration definitons, we should not end up here.
    if (model == TargetAccession and field not in TargetAccessionFields) or (
        model == TargetSequence and field not in TargetSequenceFields
    ):
        raise HTTPException(422, f"Field `{field.name}` is incompatible with target model `{model}`.")

    published_score_sets_stmt = select(ScoreSet).where(ScoreSet.published_date.is_not(None)).subquery()

    # getattr obscures MyPy errors by coercing return type to Any
    model_field = field.value.replace("-", "_")
    column_field = getattr(model, model_field)
    query = (
        select(column_field, func.count(column_field))
        .join(TargetGene)
        .group_by(column_field)
        .join_from(TargetGene, published_score_sets_stmt)
    )

    return db.execute(query).all()


# Accession based targets only.
@router.get("/target/accession/{field}", status_code=200, response_model=dict[str, int])
def target_accessions_by_field(field: TargetAccessionFields, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the provided `field` (member of the `target_accessions` table).
    Don't include any NULL field values.
    """
    return {
        field_val: count
        for field_val, count in _target_from_field_and_model(db, TargetAccession, field)
        if field_val is not None
    }


# Sequence based targets only.
@router.get("/target/sequence/{field}", status_code=200, response_model=dict[str, int])
def target_sequences_by_field(field: TargetSequenceFields, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the provided `field` (member of the `target_sequences` table).
    Don't include any NULL field values.
    """
    return {
        field_val: count
        for field_val, count in _target_from_field_and_model(db, TargetSequence, field)
        if field_val is not None
    }


# Statistics on fields relevant to both accession and sequence based targets. Generally, these require custom logic to harmonize both target sub types.
@router.get("/target/gene/{field}", status_code=200, response_model=dict[str, int])
def target_genes_by_field(field: TargetGeneFields, db: Session = Depends(get_db)) -> dict[str, int]:
    """
    Returns a dictionary of counts for the distinct values of the provided `field` (member of the `target_sequences` table).
    Don't include any NULL field values. Each field here is handled individually because of the unique structure of this
    target gene object- fields might require information from both TargetGene subtypes (accession and sequence).
    """
    association_tables: dict[TargetGeneFields, Union[type[EnsemblOffset], type[RefseqOffset], type[UniprotOffset]]] = {
        TargetGeneFields.ensemblIdentifier: EnsemblOffset,
        TargetGeneFields.refseqIdentifier: RefseqOffset,
        TargetGeneFields.uniprotIdentifier: UniprotOffset,
    }
    identifier_models: dict[
        TargetGeneFields, Union[type[EnsemblIdentifier], type[RefseqIdentifier], type[UniprotIdentifier]]
    ] = {
        TargetGeneFields.ensemblIdentifier: EnsemblIdentifier,
        TargetGeneFields.refseqIdentifier: RefseqIdentifier,
        TargetGeneFields.uniprotIdentifier: UniprotIdentifier,
    }

    # All targets linked to a published score set.
    published_score_sets_stmt = select(TargetGene).join(ScoreSet).where(ScoreSet.published_date.is_not(None)).subquery()

    # Assumes identifiers cannot be duplicated within a Target.
    if field in identifier_models.keys():
        # getattr obscures MyPy errors by coercing return type to Any
        attr_for_identifier = getattr(identifier_models[field], "identifier")

        query = (
            select(attr_for_identifier, func.count(attr_for_identifier))
            .join(association_tables[field])
            .join(published_score_sets_stmt)
            .group_by(attr_for_identifier)
        )

        return {identifier: count for identifier, count in db.execute(query).all() if identifier is not None}

    # Can't join a TargetGene query to TargetGene query, so just select the desired columns directly from the subquery.
    elif field is TargetGeneFields.category:
        query = select(published_score_sets_stmt.c.category, func.count(published_score_sets_stmt.c.category)).group_by(
            published_score_sets_stmt.c.category
        )

        return {category: count for category, count in db.execute(query).all() if category is not None}

    # Target gene organism needs special handling: it is stored differently between accession and sequence Targets.
    elif field is TargetGeneFields.organism:
        sequence_based_targets_query = (
            select(Taxonomy.organism_name, func.count(Taxonomy.organism_name))
            .join(TargetSequence)
            .join(published_score_sets_stmt)
            .group_by(Taxonomy.organism_name)
        )
        accession_based_targets_query = select(func.count(TargetAccession.id)).join(published_score_sets_stmt)

        organisms: dict[str, int] = {
            organism: count
            for organism, count in db.execute(sequence_based_targets_query).all()
            if organism is not None
        }
        accession_count = db.execute(accession_based_targets_query).scalar_one_or_none()

        # NOTE: For now (forever?), all accession based targets are human genomic sequences. It is possible this
        #       assumption changes if we add mouse (or other non-human) genomes to MaveDB.
        if "Homo sapiens" in organisms and accession_count:
            organisms["Homo sapiens"] += accession_count
        elif accession_count:
            organisms["Homo sapiens"] = accession_count

        return organisms

    # Protection from this case occurs via FastApi/pydantic Enum validation.
    else:
        raise ValueError(f"Unknown field: {field}")


def _record_from_field_and_model(
    db: Session,
    model: RecordNames,
    field: RecordFields,
):
    """
    Given a member of the RecordNames and RecordFields Enums, generate counts that can be used to create a
    statistic for those enums.

    This function should be used for generating statistics for fields shared between Experiments and Score Sets.
    If necessary, Experiment Sets can be handled in a similar manner in the future.
    """
    association_tables: dict[
        RecordNames,
        dict[
            RecordFields,
            Union[
                Table,
                type[ExperimentControlledKeywordAssociation],
                type[ExperimentPublicationIdentifierAssociation],
                type[ScoreSetPublicationIdentifierAssociation],
            ],
        ],
    ] = {
        RecordNames.experiment: {
            RecordFields.doiIdentifiers: experiments_doi_identifiers_association_table,
            RecordFields.publicationIdentifiers: ExperimentPublicationIdentifierAssociation,
            RecordFields.rawReadIdentifiers: experiments_raw_read_identifiers_association_table,
            RecordFields.keywords: ExperimentControlledKeywordAssociation,
        },
        RecordNames.scoreSet: {
            RecordFields.doiIdentifiers: score_sets_doi_identifiers_association_table,
            RecordFields.publicationIdentifiers: ScoreSetPublicationIdentifierAssociation,
            RecordFields.rawReadIdentifiers: score_sets_raw_read_identifiers_association_table,
        },
    }

    models: dict[RecordNames, Union[type[Experiment], type[ScoreSet]]] = {
        RecordNames.experiment: Experiment,
        RecordNames.scoreSet: ScoreSet,
    }

    queried_model = models[model]

    # created-by field does not operate on association tables and is defined directly on score set / experiment
    # records, so we operate directly on those records.
    # getattr obscures MyPy errors by coercing return type to Any
    model_created_by_field = getattr(queried_model, "created_by_id")
    model_published_data_field = getattr(queried_model, "published_date")
    if field is RecordFields.createdBy:
        query = (
            select(User.username, func.count(User.id))
            .join(queried_model, model_created_by_field == User.id)
            .where(model_published_data_field.is_not(None))
            .group_by(User.id)
        )

        return db.execute(query).all()
    else:
        # All assc table identifiers which are linked to a published model.
        queried_assc_table = association_tables[model][field]
        published_score_sets_statement = (
            select(queried_assc_table).join(queried_model).where(model_published_data_field.is_not(None)).subquery()
        )

    # Assumes any identifiers / keywords may not be duplicated within a record.
    if field is RecordFields.doiIdentifiers:
        query = select(DoiIdentifier.identifier, func.count(DoiIdentifier.identifier)).group_by(
            DoiIdentifier.identifier
        )
    elif field is RecordFields.keywords:
        query = select(ControlledKeyword.value, func.count(ControlledKeyword.value)).group_by(ControlledKeyword.value)
    elif field is RecordFields.rawReadIdentifiers:
        query = select(RawReadIdentifier.identifier, func.count(RawReadIdentifier.identifier)).group_by(
            RawReadIdentifier.identifier
        )

    # Handle publication identifiers separately since they may have duplicated identifiers
    elif field is RecordFields.publicationIdentifiers:
        publication_query = (
            select(
                PublicationIdentifier.identifier,
                PublicationIdentifier.db_name,
                func.count(PublicationIdentifier.identifier),
            )
            .join(published_score_sets_statement)
            .group_by(PublicationIdentifier.identifier, PublicationIdentifier.db_name)
        )

        publication_identifiers: dict[str, dict[str, int]] = {}

        for identifier, db_name, count in db.execute(publication_query).all():
            # We don't need to worry about overwriting existing identifiers within these internal dictionaries because
            # of the SQL group by clause.
            if db_name in publication_identifiers:
                publication_identifiers[db_name][identifier] = count
            else:
                publication_identifiers[db_name] = {identifier: count}

        return [(db_name, identifiers) for db_name, identifiers in publication_identifiers.items()]

    # Protection from this case occurs via FastApi/pydantic Enum validation on methods which reference this one.
    else:
        return []

    return db.execute(query.join(published_score_sets_statement)).all()


# Model based statistics for shared fields.
#
# NOTE: If custom logic is needed for record models with more specific endpoint paths,
#       i.e. non-shared fields, define them above this route so as not to obscure them.
@router.get("/record/{model}/{field}", status_code=200, response_model=Union[dict[str, int], dict[str, dict[str, int]]])
def record_object_statistics(
    model: RecordNames, field: RecordFields, db: Session = Depends(get_db)
) -> Union[dict[str, int], dict[str, dict[str, int]]]:
    """
    Resolve a dictionary of statistics based on the provided model name and model field.

    Model names and fields should be members of the Enum classes defined above. Providing an invalid model name or
    model field will yield a 422 Unprocessable Entity error with details about valid enum values.
    """
    # Validation to ensure 'keywords' is only used with 'experiment'.
    if model == RecordNames.scoreSet and field == RecordFields.keywords:
        raise HTTPException(
            status_code=422, detail="The 'keywords' field can only be used with the 'experiment' model."
        )

    count_data = _record_from_field_and_model(db, model, field)

    return {field_val: count for field_val, count in count_data if field_val is not None}



from datetime import date

from sqlalchemy import Column, Date, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, relationship

from mavedb.db.base import Base

from .score_set import ScoreSet


class Variant(Base):
    __tablename__ = "variants"

    id = Column(Integer, primary_key=True)

    urn = Column(String(64), index=True, nullable=True, unique=True)
    data = Column(JSONB, nullable=False)

    score_set_id = Column("scoreset_id", Integer, ForeignKey("scoresets.id"), index=True, nullable=False)
    # TODO examine if delete-orphan is necessary, explore cascade
    score_set: Mapped[ScoreSet] = relationship(back_populates="variants")

    hgvs_nt = Column(String, nullable=True)
    hgvs_pro = Column(String, nullable=True)
    hgvs_splice = Column(String, nullable=True)

    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)
